import unittest
import math
import numpy as np
import deepchem as dc
try:
    import torch
    has_torch = True
except ModuleNotFoundError:
    has_torch = False

if has_torch:
    from deepchem.utils import equivariance_utils


def rotation_matrix(axis, angle):
    """Generate a 3D rotation matrix."""
    axis = axis / np.linalg.norm(axis)
    a = np.cos(angle / 2.0)
    b, c, d = -axis * np.sin(angle / 2.0)
    return np.array([[
        a * a + b * b - c * c - d * d, 2 * (b * c - a * d), 2 * (b * d + a * c)
    ], [
        2 * (b * c + a * d), a * a + c * c - b * b - d * d, 2 * (c * d - a * b)
    ], [
        2 * (b * d - a * c), 2 * (c * d + a * b), a * a + d * d - b * b - c * c
    ]])


def apply_rotation(x, axis, angle):
    """Apply a 3D rotation to the positions."""
    from deepchem.utils.test.test_equivariance_utils import rotation_matrix
    R = torch.tensor(rotation_matrix(axis, angle), dtype=torch.float32)
    return torch.matmul(x, R.T)


class TestEquivarianceUtils(unittest.TestCase):
    """Test cases for the equivariance utilities."""

    def setUp(self):
        """Set up test fixtures."""
        import dgl
        from rdkit import Chem
        # Create a sample graph for testing
        # Load molecule and featurize
        mol = Chem.MolFromSmiles("CCO")
        featurizer = dc.feat.EquivariantGraphFeaturizer(fully_connected=True,
                                                        embeded=True)
        mol_graph = featurizer.featurize([mol])[0]

        self.G = dgl.graph((mol_graph.edge_index[0], mol_graph.edge_index[1]))
        self.G.ndata['f'] = torch.tensor(mol_graph.node_features,
                                         dtype=torch.float32).unsqueeze(-1)
        self.G.ndata['x'] = torch.tensor(
            mol_graph.positions, dtype=torch.float32)  # Atomic positions
        self.G.edata['d'] = torch.tensor(mol_graph.edge_features,
                                         dtype=torch.float32)
        self.G.edata['w'] = torch.tensor(mol_graph.edge_weights,
                                         dtype=torch.float32)

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_su2_generators_shape(self) -> None:
        # Test if the output has the correct shape.
        j_values = [1, 2, 3, 5,
                    7]  # Test for multiple quantum angular momentum values
        for j in j_values:
            with self.subTest(j=j):
                generators = equivariance_utils.su2_generators(j)
                self.assertEqual(generators.shape, (3, 2 * j + 1, 2 * j + 1))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_su2_generators_zero_momenta(self) -> None:
        # Test for the case of zero momentum (j=0).
        j = 0
        generators = equivariance_utils.su2_generators(j)
        expected_generators = torch.zeros((3, 1, 1), dtype=torch.complex64)
        self.assertTrue(torch.allclose(generators, expected_generators))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_su2_generators_angular_momentum_one(self) -> None:
        # Test for the case of momentum j=1 (spin-1).
        j = 1
        generators = equivariance_utils.su2_generators(j)
        # Expected J_x, J_z, J_y matrices for j=1

        expected_generators = torch.tensor(
            [[[0.0000 + 0.0000j, 0.7071 + 0.0000j, 0.0000 + 0.0000j],
              [-0.7071 + 0.0000j, 0.0000 + 0.0000j, 0.7071 + 0.0000j],
              [0.0000 + 0.0000j, -0.7071 + 0.0000j, 0.0000 + 0.0000j]],
             [[-0.0000 - 1.0000j, 0.0000 + 0.0000j, 0.0000 + 0.0000j],
              [0.0000 + 0.0000j, 0.0000 + 0.0000j, 0.0000 + 0.0000j],
              [0.0000 + 0.0000j, 0.0000 + 0.0000j, 0.0000 + 1.0000j]],
             [[0.0000 - 0.0000j, 0.0000 + 0.7071j, 0.0000 - 0.0000j],
              [0.0000 + 0.7071j, 0.0000 - 0.0000j, 0.0000 + 0.7071j],
              [0.0000 - 0.0000j, 0.0000 + 0.7071j, 0.0000 - 0.0000j]]])

        self.assertTrue(torch.allclose(generators, expected_generators))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_su2_commutation(self):
        j_values = [0, 0.5, 1, 1.5, 2,
                    2.5]  # Test for multiple quantum angular momentum values
        for j in j_values:
            with self.subTest(j=j):
                X = equivariance_utils.su2_generators(j)
                self.assertTrue(
                    torch.allclose(equivariance_utils.commutator(X[0], X[1]),
                                   X[2]))
                self.assertTrue(
                    torch.allclose(equivariance_utils.commutator(X[1], X[2]),
                                   X[0]))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_change_basis_real_to_complex_j_0(self):
        # Test for j = 0, which means we have a 1x1 transformation matrix
        j = 0
        Q = equivariance_utils.change_basis_real_to_complex(
            j, dtype=torch.complex128)
        expected_Q = torch.tensor([[1]], dtype=torch.complex128)
        self.assertTrue(torch.allclose(Q, expected_Q))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_change_basis_real_to_complex_j_2(self):
        # Test for j = 2, which means we have a 5x5 transformation matrix
        j = 2
        Q = equivariance_utils.change_basis_real_to_complex(
            j, dtype=torch.complex64)
        expected_Q = torch.tensor(
            [[
                0.0000 + 0.7071j, -0.0000 + 0.0000j, -0.0000 + 0.0000j,
                -0.0000 + 0.0000j, -0.7071 + 0.0000j
            ],
             [
                 -0.0000 + 0.0000j, 0.0000 + 0.7071j, -0.0000 + 0.0000j,
                 -0.7071 + 0.0000j, -0.0000 + 0.0000j
             ],
             [
                 -0.0000 + 0.0000j, -0.0000 + 0.0000j, -1.0000 + 0.0000j,
                 -0.0000 + 0.0000j, -0.0000 + 0.0000j
             ],
             [
                 -0.0000 + 0.0000j, 0.0000 + 0.7071j, -0.0000 + 0.0000j,
                 0.7071 - 0.0000j, -0.0000 + 0.0000j
             ],
             [
                 -0.0000 - 0.7071j, -0.0000 + 0.0000j, -0.0000 + 0.0000j,
                 -0.0000 + 0.0000j, -0.7071 + 0.0000j
             ]],
            dtype=torch.complex64)
        self.assertTrue(torch.allclose(Q, expected_Q))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_change_basis_real_to_complex_device(self) -> None:
        # Test for device placement (CPU to CUDA)
        j = 1
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        Q = equivariance_utils.change_basis_real_to_complex(j, device=device)
        self.assertEqual(Q.device, device)

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_change_basis_real_to_complex_dtype_conversion(self):
        # Test for dtype conversion (complex128 to complex64)
        j = 1
        Q_complex128 = equivariance_utils.change_basis_real_to_complex(
            j, dtype=torch.complex128)
        Q_complex64 = equivariance_utils.change_basis_real_to_complex(
            j, dtype=torch.complex64)

        self.assertEqual(Q_complex128.dtype, torch.complex128)
        self.assertEqual(Q_complex64.dtype, torch.complex64)

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_so3_generators_shape(self):
        j_values = [1, 2, 3, 4, 5]
        for j in j_values:
            with self.subTest(j=j):
                result = equivariance_utils.so3_generators(j)
                expected_shape = (3, 2 * j + 1, 2 * j + 1)
                self.assertEqual(result.shape, expected_shape)

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_so3_commutation(self):
        j_values = [0, 1, 2, 3, 4,
                    5]  # Test for multiple quantum angular momentum values
        for j in j_values:
            with self.subTest(j=j):
                X = equivariance_utils.so3_generators(j)
                self.assertTrue(
                    torch.allclose(equivariance_utils.commutator(X[0], X[1]),
                                   X[2]))
                self.assertTrue(
                    torch.allclose(equivariance_utils.commutator(X[1], X[2]),
                                   X[0]))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_so3_j0(self):
        j = 0
        result = equivariance_utils.so3_generators(j)
        expected = torch.tensor([[[0.]], [[0.]], [[0.]]],
                                dtype=torch.float64).float()
        self.assertTrue(torch.allclose(result, expected))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_so3_j1(self):
        j = 1
        result = equivariance_utils.so3_generators(j)
        expected = torch.tensor(
            [[[0.0000, 0.0000, 0.0000], [0.0000, 0.0000, -1.0000],
              [0.0000, 1.0000, 0.0000]],
             [[0.0000, 0.0000, 1.0000], [0.0000, 0.0000, 0.0000],
              [-1.0000, 0.0000, 0.0000]],
             [[0.0000, -1.0000, 0.0000], [1.0000, 0.0000, 0.0000],
              [0.0000, 0.0000, 0.0000]]],
            dtype=torch.float64).float()
        self.assertTrue(torch.allclose(result, expected, atol=1e-5))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_unitary_property(self):
        j = 2
        alpha = torch.tensor([0.2])
        beta = torch.tensor([0.1])
        gamma = torch.tensor([0.7])

        D_matrix = equivariance_utils.wigner_D(j, alpha, beta, gamma)
        D_matrix = D_matrix[0]
        conjugate_transpose = torch.transpose(torch.conj(D_matrix), 0, 1)
        identity_matrix = torch.eye(D_matrix.shape[0], dtype=D_matrix.dtype)

        self.assertTrue(
            torch.allclose(D_matrix @ conjugate_transpose,
                           identity_matrix,
                           atol=1e-5))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_orthogonality(self):
        j = 2
        alpha = torch.tensor([0.2])
        beta = torch.tensor([0.1])
        gamma = torch.tensor([0.7])

        D_matrix = equivariance_utils.wigner_D(j, alpha, beta, gamma)
        D_matrix = D_matrix[0]
        num_columns = D_matrix.shape[1]

        for col1 in range(num_columns):
            for col2 in range(num_columns):
                if col1 != col2:
                    dot_product = torch.dot(D_matrix[:, col1], D_matrix[:,
                                                                        col2])
                    self.assertAlmostEqual(dot_product.item(), 0.0, places=5)

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_semifactorial(self) -> None:
        # Test base cases
        self.assertEqual(equivariance_utils.semifactorial(0), 1.0)
        self.assertEqual(equivariance_utils.semifactorial(1), 1.0)

        self.assertEqual(equivariance_utils.semifactorial(5),
                         15.0)  # 5!! = 5 * 3 * 1
        self.assertEqual(equivariance_utils.semifactorial(6),
                         48.0)  # 6!! = 6 * 4 * 2

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_pochhammer_base_case(self) -> None:
        # Test k=0 (should return 1.0)
        self.assertEqual(equivariance_utils.pochhammer(3, 0), 1.0)

        self.assertEqual(equivariance_utils.pochhammer(3, 4),
                         360.0)  # (3)_4 = 3 * 4 * 5 * 6
        self.assertEqual(equivariance_utils.pochhammer(5, 2),
                         30.0)  # (5)_2 = 5 * 6

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_lpmv(self) -> None:
        x = torch.tensor([0.5])
        # Test P_2^1(x)
        result = equivariance_utils.lpmv(2, 1, x)
        expected = torch.tensor([-1.2990])
        self.assertTrue(torch.allclose(result, expected, atol=1e-4))

        # Test P_1^1(x)
        result = equivariance_utils.lpmv(1, 1, x)
        expected = torch.tensor([-math.sqrt(1 - 0.5**2)])
        self.assertTrue(torch.allclose(result, expected, atol=1e-4))

        # Test P_2^2(x)
        result = equivariance_utils.lpmv(2, 2, x)
        expected = torch.tensor([3 * (1 - 0.5**2)])
        self.assertTrue(torch.allclose(result, expected, atol=1e-4))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_spherical_harmonics(self) -> None:
        theta = torch.tensor([0.0, math.pi / 2])
        phi = torch.tensor([0.0, math.pi])
        sh = equivariance_utils.SphericalHarmonics()

        # Test Y_1^0(theta, phi)
        result = sh.get_element(1, 0, theta, phi)
        expected = torch.tensor([0.4886, -0.0000])
        self.assertTrue(torch.allclose(result, expected, atol=1e-4))

        # Test Y_1^1(theta, phi)
        result = sh.get_element(1, 1, theta, phi)
        expected = torch.tensor([-0.0000, 0.4886])
        self.assertTrue(torch.allclose(result, expected, atol=1e-5))

        # Get all spherical harmonics
        result = sh.get(1, theta, phi)
        expected = torch.tensor([[-0.0000, 0.4886, -0.0000],
                                 [0.0000, -0.0000, 0.4886]])
        self.assertTrue(torch.allclose(result, expected, atol=1e-4))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_irr_repr(self) -> None:
        # Test irreducible representation of SO3.
        order = 1
        alpha = 0.1
        beta = 0.2
        gamma = 0.3

        # Edge case: order = 0.
        order_zero = 0
        result_zero = equivariance_utils.irr_repr(order_zero, alpha, beta,
                                                  gamma)
        self.assertTrue(torch.allclose(result_zero, torch.tensor([[1.0]])))

        result = equivariance_utils.irr_repr(order, alpha, beta, gamma)
        expected = torch.tensor([[0.9216, 0.0587, 0.3836],
                                 [0.0198, 0.9801, -0.1977],
                                 [-0.3875, 0.1898, 0.9021]])
        self.assertTrue(torch.allclose(result, expected, atol=1e-4))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_get_matrix_kernel(self):
        # Test for computing the kernel of a matrix
        A = torch.tensor([[1.0, 2.0, 3.0], [2.0, 4.0, 6.0], [3.0, 6.0, 9.0]])
        kernel = equivariance_utils.get_matrix_kernel(A)
        for vector in kernel:
            result = torch.matmul(A, vector)
            self.assertTrue(
                torch.allclose(result, torch.zeros_like(result), atol=1e-5))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_basis_transformation_Q_J_output_shape(self):
        J, order_in, order_out = 1, 1, 1
        result = equivariance_utils.basis_transformation_Q_J(
            J, order_in, order_out)
        expected_shape = ((2 * order_out + 1) * (2 * order_in + 1), 2 * J + 1)
        self.assertEqual(result.shape, expected_shape)

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_basis_transformation_Q_J_nonzero_output(self):
        J, order_in, order_out = 1, 1, 1
        result = equivariance_utils.basis_transformation_Q_J(
            J, order_in, order_out)
        self.assertTrue(torch.any(result != 0),
                        "Output tensor should not be all zeros")

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_kron(self):
        # Test Kronecker product
        A = torch.tensor([[1.0, 2.0], [3.0, 4.0]])
        B = torch.tensor([[0.0, 5.0], [6.0, 7.0]])
        result = equivariance_utils.kron(A, B)
        expected = torch.tensor([[0.0, 5.0, 0.0, 10.0], [6.0, 7.0, 12.0, 14.0],
                                 [0.0, 15.0, 0.0, 20.0],
                                 [18.0, 21.0, 24.0, 28.0]])
        self.assertTrue(torch.allclose(result, expected, atol=1e-5))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_spherical_from_cartesian(self):
        cartesian = torch.tensor([[1.0, 1.0, 1.0]])  # [y, z, x]
        result = equivariance_utils.get_spherical_from_cartesian(cartesian)

        expected_radius = math.sqrt(1**2 + 1**2 + 1**2)
        expected_phi = math.pi / 4
        expected_theta = math.atan2(math.sqrt(1**2 + 1**2), 1)

        self.assertAlmostEqual(result[0, 0].item(), expected_radius, places=6)
        self.assertAlmostEqual(result[0, 1].item(), expected_phi, places=6)
        self.assertAlmostEqual(result[0, 2].item(), expected_theta, places=6)

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_spherical_from_cartesian_zero_vector(self):
        cartesian = torch.tensor([[0.0, 0.0, 0.0]])
        result = equivariance_utils.get_spherical_from_cartesian(cartesian)

        self.assertAlmostEqual(result[0, 0].item(), 0.0,
                               places=6)  # Radius should be 0
        self.assertAlmostEqual(result[0, 1].item(), 0.0,
                               places=6)  # Azimuth phi should be 0
        self.assertAlmostEqual(result[0, 2].item(), 0.0,
                               places=6)  # Elevation theta should be 0

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_spherical_from_cartesian_negative(self):
        cartesian = torch.tensor([[-1.0, -1.0, -1.0]])
        result = equivariance_utils.get_spherical_from_cartesian(cartesian)
        expected_radius = math.sqrt(1**2 + 1**2 + 1**1)
        expected_phi = -3 * math.pi / 4
        expected_theta = math.atan2(math.sqrt(1**2 + 1**2), -1)

        self.assertAlmostEqual(result[0, 0].item(), expected_radius, places=6)
        self.assertAlmostEqual(result[0, 1].item(), expected_phi, places=6)
        self.assertAlmostEqual(result[0, 2].item(), expected_theta, places=6)

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_spherical_from_cartesian_divide_radius_arg(self):
        cartesian = torch.tensor([[3.0, 4.0, 5.0]])
        result = equivariance_utils.get_spherical_from_cartesian(
            cartesian, divide_radius_by=2.0)
        expected_radius = math.sqrt(3**2 + 4**2 + 5**2) / 2

        self.assertAlmostEqual(result[0, 0].item(), expected_radius, places=6)

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_spherical_from_cartesian_angle_boundaries(self):
        """Ensures φ is in [-π, π] and θ is in [0, π]"""
        cartesian = torch.tensor([[2.0, -2.0, 2.0]])
        result = equivariance_utils.get_spherical_from_cartesian(cartesian)

        phi = result[0, 1].item()
        theta = result[0, 2].item()

        self.assertTrue(-math.pi <= phi <= math.pi,
                        msg=f"Azimuth φ out of bounds: {phi}")
        self.assertTrue(0 <= theta <= math.pi,
                        msg=f"Elevation θ out of bounds: {theta}")

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_spherical_from_cartesian_high_dimensional_tensor(self):
        cartesian = torch.tensor([[[1.0, 1.0, 1.0], [2.0, 2.0, 2.0]],
                                  [[0.0, 0.0, 1.0], [-1.0, -1.0, -1.0]]])
        result = equivariance_utils.get_spherical_from_cartesian(cartesian)

        self.assertEqual(result.shape, (2, 2, 3))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_precompute_sh(self):
        """Test spherical harmonics computation for a simple input."""
        r_ij = torch.tensor([[1.0, 0.5, 1.0]])  # [radius, phi, theta]
        max_J = 2
        result = equivariance_utils.precompute_sh(r_ij, max_J)

        self.assertEqual(len(result), max_J + 1)
        for J in range(max_J + 1):
            self.assertTrue(isinstance(result[J], torch.Tensor))

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_precompute_sh_zero_order(self):
        """Test when max_J = 0, only J=0 should be present."""
        r_ij = torch.tensor([[1.0, 1.0, 1.0]])
        result = equivariance_utils.precompute_sh(r_ij, max_J=0)

        self.assertEqual(len(result), 1)  # Only J=0 should be present
        self.assertIn(0, result)  # J=0 key should exist

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_precompute_sh_multiple_inputs(self):
        """Test batch inputs to check if function handles multiple r_ij vectors."""
        r_ij = torch.tensor([[1.0, 0.5, 1.0], [2.0, 1.0, 0.5]])
        max_J = 3
        result = equivariance_utils.precompute_sh(r_ij, max_J)

        self.assertEqual(len(result), max_J + 1)
        for J in range(max_J + 1):
            self.assertEqual(result[J].shape[0], r_ij.shape[0])

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_precompute_sh_output_shape(self):
        """Test if each J value has shape [B, N, K, 2J+1]."""
        r_ij = torch.tensor([[1.0, 0.5, 1.0], [2.0, 1.0, 0.5]])
        max_J = 2
        result = equivariance_utils.precompute_sh(r_ij, max_J)

        for J in range(max_J + 1):
            expected_last_dim = 2 * J + 1
            self.assertEqual(result[J].shape[-1], expected_last_dim)

    def sample_graph(self):
        """Create a test graph with SE(3) features using the SMILES 'CCO'."""
        import dgl
        from rdkit import Chem
        import deepchem as dc

        mol = Chem.MolFromSmiles("CCO")
        featurizer = dc.feat.EquivariantGraphFeaturizer(fully_connected=True,
                                                        embeded=True)
        features = featurizer.featurize([mol])[0]

        G = dgl.graph((features.edge_index[0], features.edge_index[1]),
                      num_nodes=len(features.node_features))
        G.ndata['x'] = torch.tensor(features.positions, dtype=torch.float32)
        G.edata['d'] = G.ndata['x'][G.edges()[1]] - G.ndata['x'][G.edges()[0]]

        return G

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_translation_equivariance(self):
        """Check that translating node positions does not change basis functions or relative distances."""
        from deepchem.utils.equivariance_utils import get_equivariant_basis_and_r

        basis_original, r_original = get_equivariant_basis_and_r(
            self.G, max_degree=2, compute_gradients=False)

        # Apply random translation
        translation = torch.randn(1, 3)
        G_translated = self.G.clone()
        G_translated.ndata['x'] += translation
        G_translated.edata['d'] = G_translated.ndata['x'][G_translated.edges(
        )[1]] - G_translated.ndata['x'][G_translated.edges()[0]]

        basis_translated, r_translated = get_equivariant_basis_and_r(
            G_translated, max_degree=2, compute_gradients=False)

        for key in basis_original:
            if key == '0,0':
                assert torch.allclose(
                    basis_original[key], basis_translated[key],
                    atol=1e-5), f"Failed translation equivariance for {key}"

        # r should be unchanged under translation
        assert torch.allclose(r_original, r_translated, atol=1e-5)

    @unittest.skipIf(not has_torch, "torch is not available")
    def test_rotation_equivariance(self):
        """Check that rotating node positions results in equivalent basis transformations and correct `r` behavior."""
        from deepchem.utils.equivariance_utils import get_equivariant_basis_and_r

        basis_original, r_original = get_equivariant_basis_and_r(
            self.G, max_degree=2, compute_gradients=False)

        # Apply rotation
        axis = torch.randn(3)
        angle = torch.rand(1).item() * 2 * np.pi

        G_rotated = self.G.clone()
        G_rotated.ndata['x'] = apply_rotation(self.G.ndata['x'], axis, angle)
        G_rotated.edata['d'] = G_rotated.ndata['x'][
            G_rotated.edges()[1]] - G_rotated.ndata['x'][G_rotated.edges()[0]]

        basis_rotated, r_rotated = get_equivariant_basis_and_r(
            G_rotated, max_degree=2, compute_gradients=False)

        for key in basis_original:
            if key == '0,0':
                assert torch.allclose(
                    basis_original[key], basis_rotated[key],
                    atol=1e-5), f"Failed rotation equivariance for {key}"

        # Check if the norm of `r` is unchanged under rotation
        assert torch.allclose(r_original[..., 0], r_rotated[..., 0], atol=1e-5)



"""Type annotations that are widely used in DeepChem"""

from typing import Any, Callable, List, Sequence, Tuple, TypeVar, Union, Iterable, Dict
import numpy as np

T = TypeVar("T")

# An activation function for a layer: either a function or the name of a standard activation
ActivationFn = Union[Callable, str]

# A loss function for use with KerasModel or TorchModel: f(outputs, labels, weights)
LossFn = Callable[[List, List, List], Any]

# A single value of some type, or multiple values of that type
OneOrMany = Union[T, Sequence[T]]

# The shape of a NumPy array
Shape = Tuple[int, ...]

# A NumPy array, or an object that can be converted to one.  Once we move to
# requiring NumPy 1.20, we should replace this with numpy.typing.ArrayLike.
ArrayLike = Union[np.ndarray, Sequence]

# type of RDKit object
RDKitMol = Any
RDKitAtom = Any
RDKitBond = Any

# type of Pymatgen object
PymatgenStructure = Any
PymatgenComposition = Any

Params = Union[Any, Iterable[Dict[str, Any]]]
LossClosure = Callable[[], float]
Betas2 = Tuple[float, float]
State = Dict[str, Any]
Nus2 = Tuple[float, float]



from typing import Optional, Dict, Any, List, Union, Tuple
import torch
from deepchem.utils.differentiation_utils import LinearOperator
from deepchem.utils.dft_utils import HFEngine, SpinParam, get_xc, BaseXC, SCF_QCCalc, BaseSCFEngine, BaseSystem

__all__ = ["KS"]


class KS(SCF_QCCalc):
    """
    Performing Restricted or Unrestricted Kohn-Sham DFT calculation.

    In physics and quantum chemistry, specifically density functional
    theory, the Kohn–Sham equation is the non-interacting Schrödinger
    equation (more clearly, Schrödinger-like equation) of a fictitious
    system (the "Kohn–Sham system") of non-interacting particles
    (typically electrons) that generate the same density as any given
    system of interacting particles.

    """

    def __init__(self,
                 system: BaseSystem,
                 xc: Union[str, BaseXC, None],
                 restricted: Optional[bool] = None,
                 variational: bool = False):
        """Initialize the Kohn-Sham DFT calculation.

        Parameters
        ----------
        system: BaseSystem
            System to be calculated.
        xc: Union[str, BaseXC, None]
            Exchange-correlation potential and energy to be used.
            It can accept ``None`` as an input to represent no xc
            potential involved.
        restricted: Optional[bool]
            If True, performing restricted Kohn-Sham DFT. If False,
            it performs the unrestricted Kohn-Sham DFT.
            If None, it will choose True if the system is unpolarized
            and False if it is polarized
        variational: bool
            If True, then solve the Kohn-Sham equation variationally
            (i.e. using optimization) instead of using self-consistent
            iteration. Otherwise, solve it using self-consistent iteration.

        """
        engine = KSEngine(system, xc)
        super().__init__(engine, variational)


class KSEngine(BaseSCFEngine):
    """
    Private class of Engine to be used with KS.
    This class provides the calculation of the self-consistency iteration step
    and the calculation of the post-calculation properties.

    """

    def __init__(self,
                 system: BaseSystem,
                 xc: Union[str, BaseXC, None],
                 restricted: Optional[bool] = None):
        """Initialize the Kohn-Sham DFT engine.

        Parameters
        ----------
        system: BaseSystem
            System to be calculated.
        xc: Union[str, BaseXC, None]
            Exchange-correlation potential and energy to be used.
            It can accept ``None`` as an input to represent no xc
            potential involved.
        restricted: Optional[bool]
            If True, performing restricted Kohn-Sham DFT. If False,
            it performs the unrestricted Kohn-Sham DFT.
            If None, it will choose True if the system is unpolarized
            and False if it is polarized

        """

        # get the xc object
        if isinstance(xc, str):
            self.xc: Optional[BaseXC] = get_xc(xc)
        elif isinstance(xc, BaseXC):
            self.xc = xc
        else:
            self.xc = xc

        # system = self.hf_engine.get_system()
        self._system = system

        # build and setup basis and grid
        self.hamilton = system.get_hamiltonian()
        if self.xc is not None or system.requires_grid():
            system.setup_grid()
            self.hamilton.setup_grid(system.get_grid(), self.xc)

        # get the HF engine and build the hamiltonian
        # no need to rebuild the grid because it has been constructed
        self.hf_engine = HFEngine(system,
                                  restricted=restricted,
                                  build_grid_if_necessary=False)
        self._polarized = self.hf_engine.polarized

        # get the orbital info
        self.orb_weight = system.get_orbweight(
            polarized=self._polarized)  # (norb,)
        self.norb = SpinParam.apply_fcn(
            lambda orb_weight: int(orb_weight.shape[-1]), self.orb_weight)

        # set up the vext linear operator
        self.knvext_linop = self.hamilton.get_kinnucl(
        )  # kinetic, nuclear, and external potential

    def get_system(self) -> BaseSystem:
        """Return the system object.

        Returns
        -------
        BaseSystem
            The system object.

        """
        return self._system

    @property
    def shape(self):
        """Shape of the density matrix

        Returns
        -------
        Tuple[int, int]
            Shape of the density matrix.

        """
        return self.knvext_linop.shape

    @property
    def dtype(self):
        """Dtype of the density matrix

        Returns
        -------
        torch.dtype
            Dtype of the density matrix.

        """
        return self.knvext_linop.dtype

    @property
    def device(self):
        """Device of the density matrix

        Returns
        -------
        torch.device
            Device of the density matrix.

        """
        return self.knvext_linop.device

    @property
    def polarized(self):
        """Returns if the calculation is polarized

        Returns
        -------
        bool
            If the calculation is polarized.

        """
        return self._polarized

    def dm2scp(
            self, dm: Union[torch.Tensor,
                            SpinParam[torch.Tensor]]) -> torch.Tensor:
        """Convert from density matrix to a self-consistent parameter (scp)

        Parameters
        ----------
        dm: Union[torch.Tensor, SpinParam[torch.Tensor]]
            Density matrix to be converted.

        Returns
        -------
        torch.Tensor
            Self-consistent parameter.

        """
        if isinstance(dm, torch.Tensor):  # unpolarized
            # scp is the fock matrix
            fork = self.__dm2fock(dm)
            if isinstance(fork, LinearOperator):
                return fork.fullmatrix()
            else:
                raise Exception(
                    "Unpolarized calculation should return a LinearOperator")
        else:  # polarized
            # scp is the concatenated fock matrix
            fock = self.__dm2fock(dm)
            if isinstance(fock, SpinParam):
                mat_u = fock.u.fullmatrix().unsqueeze(0)
                mat_d = fock.d.fullmatrix().unsqueeze(0)
                return torch.cat((mat_u, mat_d), dim=0)
            else:
                raise Exception(
                    "Polarized calculation should return a SpinParam of LinearOperator"
                )

    def scp2dm(
            self,
            scp: torch.Tensor) -> Union[torch.Tensor, SpinParam[torch.Tensor]]:
        """Convert from self-consistent parameter (scp) to density matrix

        Parameters
        ----------
        scp: torch.Tensor
            Self-consistent parameter to be converted.

        Returns
        -------
        Union[torch.Tensor, SpinParam[torch.Tensor]]
            Density matrix.

        """
        return self.hf_engine.scp2dm(scp)

    def scp2scp(self, scp: torch.Tensor) -> torch.Tensor:
        """Self-consistent iteration step from a self-consistent parameter (scp)
        to an scp

        Parameters
        ----------
        scp: torch.Tensor
            Self-consistent parameter to be converted.

        Returns
        -------
        torch.Tensor
            New self-consistent parameter.

        """
        dm = self.scp2dm(scp)
        return self.dm2scp(dm)

    def aoparams2ene(self,
                     aoparams: torch.Tensor,
                     aocoeffs: torch.Tensor,
                     with_penalty: Optional[float] = None) -> torch.Tensor:
        """Calculate the energy from the atomic orbital params

        Parameters
        ----------
        aoparams: torch.Tensor
            Atomic orbital parameters.
        aocoeffs: torch.Tensor
            Atomic orbital coefficients.
        with_penalty: Optional[float]
            Penalty factor to be added to the energy.

        Returns
        -------
        torch.Tensor
            Energy value.

        """
        dm, penalty = self.aoparams2dm(aoparams, aocoeffs, with_penalty)
        ene = self.dm2energy(dm)
        return (ene + penalty) if penalty is not None else ene

    def aoparams2dm(self, aoparams: torch.Tensor, aocoeffs: torch.Tensor,
                    with_penalty: Optional[float] = None) -> \
            Tuple[Union[torch.Tensor, SpinParam[torch.Tensor]], Optional[torch.Tensor]]:
        """Convert the aoparams to density matrix and penalty factor

        Parameters
        ----------
        aoparams: torch.Tensor
            Atomic orbital parameters.
        aocoeffs: torch.Tensor
            Atomic orbital coefficients.
        with_penalty: Optional[float]
            Penalty factor to be added to the energy.

        Returns
        -------
        Tuple[Union[torch.Tensor, SpinParam[torch.Tensor]], Optional[torch.Tensor]]
            Density matrix and the penalty factor.

        """
        return self.hf_engine.aoparams2dm(aoparams, aocoeffs, with_penalty)

    def pack_aoparams(
            self, aoparams: Union[torch.Tensor,
                                  SpinParam[torch.Tensor]]) -> torch.Tensor:
        """Check if polarized, then pack it by concatenating them in the last dimension

        Parameters
        ----------
        aoparams: Union[torch.Tensor, SpinParam[torch.Tensor]]
            Atomic orbital parameters.

        Returns
        -------
        torch.Tensor
            Packed atomic orbital parameters.

        """
        return self.hf_engine.pack_aoparams(aoparams)

    def unpack_aoparams(
            self, aoparams: torch.Tensor
    ) -> Union[torch.Tensor, SpinParam[torch.Tensor]]:
        """Check if polarized, then construct the SpinParam (reverting the pack_aoparams)

        Parameters
        ----------
        aoparams: torch.Tensor
            Packed atomic orbital parameters.

        Returns
        -------
        Union[torch.Tensor, SpinParam[torch.Tensor]]
            Atomic orbital parameters.

        """
        return self.hf_engine.unpack_aoparams(aoparams)

    def set_eigen_options(self, eigen_options: Dict[str, Any]) -> None:
        """Set the eigendecomposition (diagonalization) option

        Parameters
        ----------
        eigen_options: Dict[str, Any]
            Options for the eigendecomposition.

        """
        self.hf_engine.set_eigen_options(eigen_options)

    def dm2energy(
            self, dm: Union[torch.Tensor,
                            SpinParam[torch.Tensor]]) -> torch.Tensor:
        """Calculate the energy given the density matrix

        Parameters
        ----------
        dm: Union[torch.Tensor, SpinParam[torch.Tensor]]
            Density matrix.

        Returns
        -------
        torch.Tensor
            Energy.

        """
        dmtot = SpinParam.sum(dm)
        e_core = self.hamilton.get_e_hcore(dmtot)
        e_elrep = self.hamilton.get_e_elrep(dmtot)
        if self.xc is not None:
            e_xc: Union[torch.Tensor, float] = self.hamilton.get_e_xc(dm)
        else:
            e_xc = 0.0
        return e_core + e_elrep + e_xc + self._system.get_nuclei_energy()

    def __dm2fock(
        self, dm: Union[torch.Tensor, SpinParam[torch.Tensor]]
    ) -> Union[LinearOperator, SpinParam[LinearOperator]]:
        """From density matrix, returns the fock matrix

        Parameters
        ----------
        dm: Union[torch.Tensor, SpinParam[torch.Tensor]]
            Density matrix.

        Returns
        -------
        Union[LinearOperator, SpinParam[LinearOperator]]
            Fock matrix.

        """
        elrep = self.hamilton.get_elrep(SpinParam.sum(dm))  # (..., nao, nao)
        core_coul = self.knvext_linop + elrep

        if self.xc is not None:
            vxc = self.hamilton.get_vxc(
                dm)  # spin param or tensor (..., nao, nao)
            return SpinParam.apply_fcn(lambda vxc_: vxc_ + core_coul, vxc)
        else:
            if isinstance(dm, SpinParam):
                return SpinParam(u=core_coul, d=core_coul)
            else:
                return core_coul

    def getparamnames(self, methodname: str, prefix: str = "") -> List[str]:
        """Parameter names for the given method

        Parameters
        ----------
        methodname: str
            Method name.
        prefix: str
            Prefix to be added to the parameter names.

        Returns
        -------
        List[str]
            Parameter names.

        """
        if methodname == "scp2scp":
            return self.getparamnames("scp2dm", prefix=prefix) + \
                self.getparamnames("dm2scp", prefix=prefix)
        elif methodname == "scp2dm":
            return self.hf_engine.getparamnames("scp2dm",
                                                prefix=prefix + "hf_engine.")
        elif methodname == "dm2scp":
            return self.getparamnames("__dm2fock", prefix=prefix)
        elif methodname == "aoparams2ene":
            return self.getparamnames("aoparams2dm", prefix=prefix) + \
                self.getparamnames("dm2energy", prefix=prefix)
        elif methodname in ["aoparams2dm", "pack_aoparams", "unpack_aoparams"]:
            return self.hf_engine.getparamnames(methodname,
                                                prefix=prefix + "hf_engine.")
        elif methodname == "dm2energy":
            hprefix = prefix + "hamilton."
            sprefix = prefix + "_system."

            if self.xc is not None:
                e_xc_params = self.hamilton.getparamnames("get_e_xc",
                                                          prefix=hprefix)
            else:
                e_xc_params = []

            return self.hamilton.getparamnames("get_e_hcore", prefix=hprefix) + \
                self.hamilton.getparamnames("get_e_elrep", prefix=hprefix) + \
                e_xc_params + \
                self._system.getparamnames("get_nuclei_energy", prefix=sprefix)
        elif methodname == "__dm2fock":
            hprefix = prefix + "hamilton."

            if self.xc is not None:
                vxc_params = self.hamilton.getparamnames("get_vxc",
                                                         prefix=hprefix)
            else:
                vxc_params = []

            return self.hamilton.getparamnames("get_elrep", prefix=hprefix) + \
                vxc_params + \
                self.knvext_linop._getparamnames(prefix=prefix + "knvext_linop.")
        else:
            raise KeyError("Method %s has no paramnames set" % methodname)



from deepchem.models.torch_models.hf_models import HuggingFaceModel
from transformers import AutoTokenizer, AutoModelForMaskedLM, AutoConfig, AutoModelForSequenceClassification
try:
    import torch
    has_torch = True
except:
    has_torch = False


class MoLFormer(HuggingFaceModel):
    """
    MoLFormer is a large-scale chemical language model designed with the intention of learning a model trained
    on small molecules which are represented as SMILES strings. MoLFormer leverges masked language modeling
    and employs a linear attention Transformer combined with rotary embeddings.

    MoLFormer-XL-both-10pct is the version trained on 10% ZINC + 10% PubChem.

    Parameters
    ----------
    task: str
        The task defines the type of learning task in the model. The supported tasks are
         - `mlm` - masked language modeling commonly used in pretraining
         - `mtr` - multitask regression - a task used for both pretraining base models and finetuning
         - `regression` - use it for regression tasks, like property prediction
         - `classification` - use it for classification tasks
    tokenizer_path: str
        Path containing pretrained tokenizer used to tokenize SMILES string for model inputs. The tokenizer path can either be a huggingFace tokenizer model or a path in the local machine containing the tokenizer.
    n_tasks: int, default 1
        Number of prediction targets for a multitask learning model

    Example
    -------
    >>> import os
    >>> import tempfile
    >>> tempdir = tempfile.mkdtemp()

    >>> # preparing dataset
    >>> import pandas as pd
    >>> import deepchem as dc
    >>> smiles = ["CCN(CCSC)C(=O)N[C@@](C)(CC)C(F)(F)F","CC1(C)CN(C(=O)Nc2cc3ccccc3nn2)C[C@@]2(CCOC2)O1"]
    >>> labels = [3.112,2.432]
    >>> df = pd.DataFrame(list(zip(smiles, labels)), columns=["smiles", "task1"])
    >>> with dc.utils.UniversalNamedTemporaryFile(mode='w') as tmpfile:
    ...     df.to_csv(tmpfile.name)
    ...     loader = dc.data.CSVLoader(["task1"], feature_field="smiles", featurizer=dc.feat.DummyFeaturizer())
    ...     dataset = loader.create_dataset(tmpfile.name)

    >>> # pretraining
    >>> from deepchem.models.torch_models.molformer import MoLFormer
    >>> pretrain_model_dir = os.path.join(tempdir, 'pretrain-molformer-model')
    >>> tokenizer_path = "ibm/MoLFormer-XL-both-10pct"
    >>> pretrain_model = MoLFormer(task='mlm', model_dir=pretrain_model_dir, tokenizer_path=tokenizer_path)  # mlm pretraining
    >>> pretraining_loss = pretrain_model.fit(dataset, nb_epoch=1)

    >>> # finetuning in regression mode
    >>> finetune_model_dir = os.path.join(tempdir, 'finetune-model')
    >>> finetune_model = MoLFormer(task='regression', model_dir=finetune_model_dir, tokenizer_path=tokenizer_path)
    >>> finetune_model.load_from_pretrained(pretrain_model_dir)
    >>> finetuning_loss = finetune_model.fit(dataset, nb_epoch=1)

    >>> # prediction and evaluation
    >>> result = finetune_model.predict(dataset)
    >>> eval_results = finetune_model.evaluate(dataset, metrics=dc.metrics.Metric(dc.metrics.mae_score))


    Reference
    ---------
    .. Ross, Jerret and Belgodere, Brian and Chenthamarakshan, Vijil and Padhi, Inkit and Mroueh, Youssef and
    .. Das, Payel "Large-Scale Chemical Language Representations Capture Molecular Structure and Properties"
    .. https://doi.org/10.48550/arxiv.2106.09553
    """

    def __init__(self,
                 task: str,
                 tokenizer_path: str = 'ibm/MoLFormer-XL-both-10pct',
                 n_tasks: int = 1,
                 **kwargs):
        self.n_tasks = n_tasks
        tokenizer = AutoTokenizer.from_pretrained(tokenizer_path,
                                                  trust_remote_code=True)
        molformer_config = AutoConfig.from_pretrained(
            "ibm/MoLFormer-XL-both-10pct", trust_remote_code=True)
        if task == 'mlm':
            model = AutoModelForMaskedLM.from_config(molformer_config,
                                                     trust_remote_code=True)
        elif task == 'mtr':
            molformer_config.problem_type = 'regression'
            molformer_config.num_labels = n_tasks
            model = AutoModelForSequenceClassification.from_config(
                config=molformer_config, trust_remote_code=True)
        elif task == 'regression':
            molformer_config.problem_type = 'regression'
            molformer_config.num_labels = n_tasks
            model = AutoModelForSequenceClassification.from_config(
                config=molformer_config, trust_remote_code=True)
        elif task == 'classification':
            if n_tasks == 1:
                molformer_config.problem_type = 'single_label_classification'
            else:
                molformer_config.num_labels = n_tasks
                molformer_config.problem_type = 'multi_label_classification'
            model = AutoModelForSequenceClassification.from_config(
                molformer_config, trust_remote_code=True)
        else:
            raise ValueError('invalid task specification')

        super(MoLFormer, self).__init__(model=model,
                                        task=task,
                                        tokenizer=tokenizer,
                                        **kwargs)

    def _prepare_batch(self, batch):
        """
        Prepares a batch of data for the model based on the specified task. It overrides the _prepare_batch
        of parent class for the following condition:-

        - When n_task == 1 and task == 'classification', CrossEntropyLoss is used which takes input in
        long int format.
        - When n_task > 1 and task == 'classification', BCEWithLogitsLoss is used which takes input in
        float format.
        """
        smiles_batch, y, w = batch
        tokens = self.tokenizer(smiles_batch[0].tolist(),
                                padding=True,
                                return_tensors="pt")

        if self.task == 'mlm':
            inputs, labels = self.data_collator.torch_mask_tokens(
                tokens['input_ids'])
            inputs = {
                'input_ids': inputs.to(self.device),
                'labels': labels.to(self.device),
                'attention_mask': tokens['attention_mask'].to(self.device),
            }
            return inputs, None, w
        elif self.task in ['regression', 'classification', 'mtr']:
            if y is not None:
                # y is None during predict
                y = torch.from_numpy(y[0])
                if self.task == 'regression' or self.task == 'mtr':
                    y = y.float().to(self.device)
                elif self.task == 'classification':
                    if self.n_tasks == 1:
                        y = y.long().to(self.device)
                    else:
                        y = y.float().to(self.device)
            for key, value in tokens.items():
                tokens[key] = value.to(self.device)

            inputs = {**tokens, 'labels': y}
            return inputs, y, w



"""Optimizers and related classes for use with TensorGraph."""

import math
from functools import partial
try:
    from deepchem.utils.optimizer_utils import LambOptimizer
except:
    pass
from typing import Dict, Union, Optional


class Optimizer(object):
    """An algorithm for optimizing a model.

    This is an abstract class.  Subclasses represent specific optimization algorithms.
    """

    def __init__(self, learning_rate: "Union[float, LearningRateSchedule]"):
        """This constructor should only be called by subclasses.

        Parameters
        ----------
        learning_rate: float or LearningRateSchedule
            the learning rate to use for optimization
        """
        self.learning_rate = learning_rate

    def _create_tf_optimizer(self, global_step):
        """Construct a TensorFlow optimizer.

        Parameters
        ----------
        global_step: tensor
            a tensor containing the global step index during optimization, used for learning rate decay

        Returns
        -------
        a new TensorFlow optimizer implementing the algorithm
        """
        raise NotImplementedError("Subclasses must implement this")

    def _create_pytorch_optimizer(self, params):
        """Construct a PyTorch optimizer.

        Parameters
        ----------
        params: Iterable
            the model parameters to optimize

        Returns
        -------
        a new PyTorch optimizer implementing the algorithm
        """
        raise NotImplementedError("Subclasses must implement this")

    def _create_jax_optimizer(self):
        """Construct a Jax optimizer.

        Returns
        -------
        a new Optax optimizer optax.GradientTransformation implementing the algorithm
        """
        raise NotImplementedError("Subclasses must implement this")


class LearningRateSchedule(object):
    """A schedule for changing the learning rate over the course of optimization.

    This is an abstract class.  Subclasses represent specific schedules.
    """

    def _create_tf_tensor(self, global_step):
        """Construct a tensor that equals the learning rate.

        Parameters
        ----------
        global_step: tensor
            a tensor containing the global step index during optimization

        Returns
        -------
        a tensor that equals the learning rate
        """
        raise NotImplementedError("Subclasses must implement this")

    def _create_pytorch_schedule(self, optimizer):
        """Construct a PyTorch learning rate scheduler.

        Parameters
        ----------
        optimizer: torch.optim.Optimizer
            the Optimizer whose learning rate will be modified

        Returns
        -------
        a PyTorch scheduler implementing the schedule
        """
        raise NotImplementedError("Subclasses must implement this")

    def _create_jax_schedule(self, learning_rate):
        """Construct a Jax learning rate scheduler using optax.

        Parameters
        ----------
        learning_rate: float
            the initial learning rate that will be modified

        Returns
        -------
        a optax scheduler implementing the schedule
        """
        raise NotImplementedError("Subclasses must implement this")


class AdaGrad(Optimizer):
    """The AdaGrad optimization algorithm.

    Adagrad is an optimizer with parameter-specific learning rates, which are
    adapted relative to how frequently a parameter gets updated during training.
    The more updates a parameter receives, the smaller the updates. See [1]_ for
    a full reference for the algorithm.

    References
    ----------
    .. [1] Duchi, John, Elad Hazan, and Yoram Singer. "Adaptive subgradient
        methods for online learning and stochastic optimization." Journal of machine
        learning research 12.7 (2011).
    """

    def __init__(self,
                 learning_rate: Union[float, LearningRateSchedule] = 0.001,
                 initial_accumulator_value: float = 0.1,
                 epsilon: float = 1e-07):
        """Construct an AdaGrad optimizer.
        Parameters
        ----------
        learning_rate: float or LearningRateSchedule
            the learning rate to use for optimization
        initial_accumulator_value: float
            a parameter of the AdaGrad algorithm
        epsilon: float
            a parameter of the AdaGrad algorithm

        """
        super(AdaGrad, self).__init__(learning_rate)
        self.initial_accumulator_value = initial_accumulator_value
        self.epsilon = epsilon

    def _create_tf_optimizer(self, global_step):
        import tensorflow as tf
        if isinstance(self.learning_rate, LearningRateSchedule):
            learning_rate = self.learning_rate._create_tf_tensor(global_step)
        else:
            learning_rate = self.learning_rate
        return tf.keras.optimizers.legacy.Adagrad(
            learning_rate=learning_rate,
            initial_accumulator_value=self.initial_accumulator_value,
            epsilon=self.epsilon)

    def _create_pytorch_optimizer(self, params):
        import torch
        if isinstance(self.learning_rate, LearningRateSchedule):
            lr = self.learning_rate.initial_rate
        else:
            lr = self.learning_rate
        return torch.optim.Adagrad(
            params,
            lr,
            initial_accumulator_value=self.initial_accumulator_value,
            eps=self.epsilon)

    def _create_jax_optimizer(self):
        import optax
        process = []
        if isinstance(self.learning_rate, LearningRateSchedule):
            lr = self.learning_rate.initial_rate
            last_process = optax.scale(-1.0)
        else:
            lr = self.learning_rate
            last_process = optax.scale(-1.0 * lr)

        process.append(
            optax.scale_by_rss(
                initial_accumulator_value=self.initial_accumulator_value,
                eps=self.epsilon))
        process.append(last_process)
        return optax.chain(*process)


class Adam(Optimizer):
    """The Adam optimization algorithm."""

    def __init__(self,
                 learning_rate: Union[float, LearningRateSchedule] = 0.001,
                 beta1: float = 0.9,
                 beta2: float = 0.999,
                 epsilon: float = 1e-08,
                 weight_decay: float = 0):
        """Construct an Adam optimizer.

        Parameters
        ----------
        learning_rate: float or LearningRateSchedule
            the learning rate to use for optimization
        beta1: float
            a parameter of the Adam algorithm
        beta2: float
            a parameter of the Adam algorithm
        epsilon: float
            a parameter of the Adam algorithm
        weight_decay: float
            L2 penalty - a parameter of the Adam algorithm
        """
        super(Adam, self).__init__(learning_rate)
        self.beta1 = beta1
        self.beta2 = beta2
        self.epsilon = epsilon
        self.weight_decay = weight_decay

    def _create_tf_optimizer(self, global_step):
        import tensorflow as tf
        if isinstance(self.learning_rate, LearningRateSchedule):
            learning_rate = self.learning_rate._create_tf_tensor(global_step)
        else:
            learning_rate = self.learning_rate
        return tf.keras.optimizers.legacy.Adam(learning_rate=learning_rate,
                                               beta_1=self.beta1,
                                               beta_2=self.beta2,
                                               epsilon=self.epsilon)

    def _create_pytorch_optimizer(self, params):
        import torch
        if isinstance(self.learning_rate, LearningRateSchedule):
            lr = self.learning_rate.initial_rate
        else:
            lr = self.learning_rate
        return torch.optim.Adam(params,
                                lr=lr,
                                betas=(self.beta1, self.beta2),
                                eps=self.epsilon,
                                weight_decay=self.weight_decay)

    def _create_jax_optimizer(self):
        import optax
        process = []
        if isinstance(self.learning_rate, LearningRateSchedule):
            scheduler = self.learning_rate._create_jax_schedule()
            process.append(optax.scale_by_schedule(scheduler))
            last_process = optax.scale(-1.0)
        else:
            lr = self.learning_rate
            last_process = optax.scale(-1.0 * lr)

        process.append(
            optax.scale_by_adam(b1=self.beta1, b2=self.beta2, eps=self.epsilon))
        process.append(last_process)
        return optax.chain(*process)


class SparseAdam(Optimizer):
    """The Sparse Adam optimization algorithm, also known as Lazy Adam.
    Sparse Adam is suitable for sparse tensors. It handles sparse updates more efficiently.
    It only updates moving-average accumulators for sparse variable indices that appear in the current batch, rather than updating the accumulators for all indices.
    """

    def __init__(self,
                 learning_rate: Union[float, LearningRateSchedule] = 0.001,
                 beta1: float = 0.9,
                 beta2: float = 0.999,
                 epsilon: float = 1e-08):
        """Construct an Adam optimizer.

        Parameters
        ----------
        learning_rate: float or LearningRateSchedule
            the learning rate to use for optimization
        beta1: float
            a parameter of the SparseAdam algorithm
        beta2: float
            a parameter of the SparseAdam algorithm
        epsilon: float
            a parameter of the SparseAdam algorithm
        """
        super(SparseAdam, self).__init__(learning_rate)
        self.beta1 = beta1
        self.beta2 = beta2
        self.epsilon = epsilon

    def _create_tf_optimizer(self, global_step):
        import tensorflow_addons as tfa
        if isinstance(self.learning_rate, LearningRateSchedule):
            learning_rate = self.learning_rate._create_tf_tensor(global_step)
        else:
            learning_rate = self.learning_rate
        return tfa.optimizers.LazyAdam(learning_rate=learning_rate,
                                       beta_1=self.beta1,
                                       beta_2=self.beta2,
                                       epsilon=self.epsilon)

    def _create_pytorch_optimizer(self, params):
        import torch
        if isinstance(self.learning_rate, LearningRateSchedule):
            lr = self.learning_rate.initial_rate
        else:
            lr = self.learning_rate
        return torch.optim.SparseAdam(params, lr, (self.beta1, self.beta2),
                                      self.epsilon)


class AdamW(Optimizer):
    """The AdamW optimization algorithm.
    AdamW is a variant of Adam, with improved weight decay.
    In Adam, weight decay is implemented as: weight_decay (float, optional) – weight decay (L2 penalty) (default: 0)
    In AdamW, weight decay is implemented as: weight_decay (float, optional) – weight decay coefficient (default: 1e-2)
    """

    def __init__(self,
                 learning_rate: Union[float, LearningRateSchedule] = 0.001,
                 weight_decay: Union[float, LearningRateSchedule] = 0.01,
                 beta1: float = 0.9,
                 beta2: float = 0.999,
                 epsilon: float = 1e-08,
                 amsgrad: bool = False):
        """Construct an AdamW optimizer.
        Parameters
        ----------
        learning_rate: float or LearningRateSchedule
            the learning rate to use for optimization
        weight_decay: float or LearningRateSchedule
            weight decay coefficient for AdamW
        beta1: float
            a parameter of the Adam algorithm
        beta2: float
            a parameter of the Adam algorithm
        epsilon: float
            a parameter of the Adam algorithm
        amsgrad: bool
            If True, will use the AMSGrad variant of AdamW (from "On the Convergence of Adam and Beyond"), else will use the original algorithm.
        """
        super(AdamW, self).__init__(learning_rate)
        self.weight_decay = weight_decay
        self.beta1 = beta1
        self.beta2 = beta2
        self.epsilon = epsilon
        self.amsgrad = amsgrad

    def _create_tf_optimizer(self, global_step):
        import tensorflow_addons as tfa
        if isinstance(self.learning_rate, LearningRateSchedule):
            learning_rate = self.learning_rate._create_tf_tensor(global_step)
        else:
            learning_rate = self.learning_rate
        return tfa.optimizers.AdamW(weight_decay=self.weight_decay,
                                    learning_rate=learning_rate,
                                    beta_1=self.beta1,
                                    beta_2=self.beta2,
                                    epsilon=self.epsilon,
                                    amsgrad=self.amsgrad)

    def _create_pytorch_optimizer(self, params):
        import torch
        if isinstance(self.learning_rate, LearningRateSchedule):
            lr = self.learning_rate.initial_rate
        else:
            lr = self.learning_rate
        return torch.optim.AdamW(params, lr, (self.beta1, self.beta2),
                                 self.epsilon, self.weight_decay, self.amsgrad)

    def _create_jax_optimizer(self):
        import optax
        process = []
        if isinstance(self.learning_rate, LearningRateSchedule):
            scheduler = self.learning_rate._create_jax_schedule()
            process.append(optax.scale_by_schedule(scheduler))
            last_process = optax.scale(-1.0)
        else:
            lr = self.learning_rate
            last_process = optax.scale(-1.0 * lr)

        process.append(
            optax.scale_by_adam(b1=self.beta1,
                                b2=self.beta2,
                                eps=self.epsilon,
                                eps_root=0.0))
        process.append(optax.add_decayed_weights(self.weight_decay, None))
        process.append(last_process)
        return optax.chain(*process)


class RMSProp(Optimizer):
    """RMSProp Optimization algorithm."""

    def __init__(self,
                 learning_rate: Union[float, LearningRateSchedule] = 0.001,
                 momentum: float = 0.0,
                 decay: float = 0.9,
                 epsilon: float = 1e-10):
        """Construct an RMSProp Optimizer.

        Parameters
        ----------
        learning_rate: float or LearningRateSchedule
            the learning_rate used for optimization
        momentum: float, default 0.0
            a parameter of the RMSProp algorithm
        decay: float, default 0.9
            a parameter of the RMSProp algorithm
        epsilon: float, default 1e-10
            a parameter of the RMSProp algorithm
        """
        super(RMSProp, self).__init__(learning_rate)
        self.momentum = momentum
        self.decay = decay
        self.epsilon = epsilon

    def _create_tf_optimizer(self, global_step):
        import tensorflow as tf
        if isinstance(self.learning_rate, LearningRateSchedule):
            learning_rate = self.learning_rate._create_tf_tensor(global_step)
        else:
            learning_rate = self.learning_rate
        return tf.keras.optimizers.legacy.RMSprop(learning_rate=learning_rate,
                                                  momentum=self.momentum,
                                                  rho=self.decay,
                                                  epsilon=self.epsilon)

    def _create_pytorch_optimizer(self, params):
        import torch
        if isinstance(self.learning_rate, LearningRateSchedule):
            lr = self.learning_rate.initial_rate
        else:
            lr = self.learning_rate
        return torch.optim.RMSprop(params,
                                   lr,
                                   alpha=self.decay,
                                   eps=self.epsilon,
                                   momentum=self.momentum)

    def _create_jax_optimizer(self):
        import optax
        process = []
        if isinstance(self.learning_rate, LearningRateSchedule):
            scheduler = self.learning_rate._create_jax_schedule()
            process.append(optax.scale_by_schedule(scheduler))
            last_process = optax.scale(-1.0)
        else:
            lr = self.learning_rate
            last_process = optax.scale(-1.0 * lr)

        process.append(
            optax.scale_by_rms(decay=self.decay,
                               eps=self.epsilon,
                               initial_scale=0.0))
        if self.momentum is not None or self.momentum != 0.0:
            process.append(optax.trace(decay=self.momentum, nesterov=False))
        process.append(last_process)
        return optax.chain(*process)


class GradientDescent(Optimizer):
    """The gradient descent optimization algorithm."""

    def __init__(self,
                 learning_rate: Union[float, LearningRateSchedule] = 0.001):
        """Construct a gradient descent optimizer.

        Parameters
        ----------
        learning_rate: float or LearningRateSchedule
            the learning rate to use for optimization
        """
        super(GradientDescent, self).__init__(learning_rate)

    def _create_tf_optimizer(self, global_step):
        import tensorflow as tf
        if isinstance(self.learning_rate, LearningRateSchedule):
            learning_rate = self.learning_rate._create_tf_tensor(global_step)
        else:
            learning_rate = self.learning_rate
        return tf.keras.optimizers.legacy.SGD(learning_rate=learning_rate)

    def _create_pytorch_optimizer(self, params):
        import torch
        if isinstance(self.learning_rate, LearningRateSchedule):
            lr = self.learning_rate.initial_rate
        else:
            lr = self.learning_rate
        return torch.optim.SGD(params, lr)

    def _create_jax_optimizer(self):
        import optax
        process = []
        if isinstance(self.learning_rate, LearningRateSchedule):
            scheduler = self.learning_rate._create_jax_schedule()
            process.append(optax.scale_by_schedule(scheduler))
            last_process = optax.scale(-1.0)
        else:
            lr = self.learning_rate
            last_process = optax.scale(-1.0 * lr)
        process.append(last_process)
        return optax.chain(*process)


class ExponentialDecay(LearningRateSchedule):
    """A learning rate that decreases exponentially with the number of training steps."""

    def __init__(self,
                 initial_rate: float,
                 decay_rate: float,
                 decay_steps: int,
                 staircase: bool = True):
        """Create an exponentially decaying learning rate.

        The learning rate starts as initial_rate.  Every decay_steps training steps, it is multiplied by decay_rate.

        Parameters
        ----------
        initial_rate: float
            the initial learning rate
        decay_rate: float
            the base of the exponential
        decay_steps: int
            the number of training steps over which the rate decreases by decay_rate
        staircase: bool
            if True, the learning rate decreases by discrete jumps every decay_steps.
            if False, the learning rate decreases smoothly every step
        """
        self.initial_rate = initial_rate
        self.decay_rate = decay_rate
        self.decay_steps = decay_steps
        self.staircase = staircase

    def _create_tf_tensor(self, global_step):
        import tensorflow as tf
        return tf.keras.optimizers.schedules.ExponentialDecay(
            initial_learning_rate=self.initial_rate,
            decay_rate=self.decay_rate,
            decay_steps=self.decay_steps,
            staircase=self.staircase)(global_step)

    def _create_pytorch_schedule(self, optimizer):
        import torch
        if self.staircase:
            return torch.optim.lr_scheduler.StepLR(optimizer, self.decay_steps,
                                                   self.decay_rate)
        return torch.optim.lr_scheduler.ExponentialLR(
            optimizer, math.pow(self.decay_rate, 1 / self.decay_steps))

    def _create_jax_schedule(self):
        import optax
        return optax.exponential_decay(init_value=self.initial_rate,
                                       transition_steps=self.decay_steps,
                                       decay_rate=self.decay_rate,
                                       staircase=self.staircase)


class LambdaLRWithWarmup(LearningRateSchedule):
    """A learning rate scheduler supporting warmup followed by cool down.

    Example
    -------
    >>> import torch
    >>> opt = Adam(learning_rate=5e-5)
    >>> lr_schedule = LambdaLRWithWarmup(initial_rate=5e-5,
    ...     num_training_steps=100, num_warmup_steps=10)
    >>> params = [torch.nn.Parameter(torch.Tensor([1.0]))]
    >>> optimizer = opt._create_pytorch_optimizer(params)
    >>> scheduler = lr_schedule._create_pytorch_schedule(optimizer)
    """

    def __init__(self,
                 initial_rate: float,
                 num_warmup_steps: int,
                 num_training_steps: Optional[int] = None,
                 warmup_type: str = 'linear'):
        """
        Parameters
        ----------
        initial_rate: float
            Initial learning rate
        num_warmup_steps: int
            Number of warmup steps
        num_training_steps: int
            Number of training steps - required for linear schedule.
        warmup_type: str, optional. default: linear
            When `linear`, creates a learning rate schedule that decreases linearly from
                the initial lr in the optimizer to 0.
            When `constant`, creates a constant learning rate preceded by a warmup period
                during which the learning rate increases linearly between 0 and the initial
                lr set in the optimizer.
        """
        assert warmup_type == 'linear' or 'constant', f'Warmup type {warmup_type} is not supported.'
        self.initial_rate = initial_rate
        self.num_warmup_steps = num_warmup_steps
        self.num_training_steps = num_training_steps
        self.warmup_type = warmup_type

    def _create_pytorch_schedule(self, optimizer):
        """Creates a PyTorch learning rate scheduler for the given optimizer.

        When the warmup type is linear, the method _linear_schedule_with_warmup
        is used to create a learning rate schedule such that the learning
        rate increases linearly from 0 to initial_rate and
        then cools down to 0 linearly.

        When the warmup type is constant, the method _constant_schedule_with_warmup
        is used to create a learning rate schedule such that the learning
        rate linearly increases form 0 to initial_rate and then stays constant.
        """

        def _linear_schedule_with_warmup(current_step: int, *,
                                         num_warmup_steps: int,
                                         num_training_steps: int):
            if current_step < num_warmup_steps:
                return float(current_step) / float(max(1, num_warmup_steps))
            return max(
                0.0,
                float(num_training_steps - current_step) /
                float(max(1, num_training_steps - num_warmup_steps)))

        def _constant_schedule_with_warmup(current_step: int, *,
                                           num_warmup_steps: int):
            if current_step < num_warmup_steps:
                return float(current_step) / float(max(1.0, num_warmup_steps))
            return 1.0

        if self.warmup_type == 'linear':
            f = partial(_linear_schedule_with_warmup,
                        num_warmup_steps=self.num_warmup_steps,
                        num_training_steps=self.num_training_steps)
        elif self.warmup_type == 'constant':
            f = partial(_constant_schedule_with_warmup,
                        num_warmup_steps=self.num_warmup_steps)

        import torch
        return torch.optim.lr_scheduler.LambdaLR(optimizer, f)


class PolynomialDecay(LearningRateSchedule):
    """A learning rate that decreases from an initial value to a final value over a fixed number of training steps."""

    def __init__(self,
                 initial_rate: float,
                 final_rate: float,
                 decay_steps: int,
                 power: float = 1.0):
        """Create a smoothly decaying learning rate.

        The learning rate starts as initial_rate.  It smoothly decreases to final_rate over decay_steps training steps.
        It decays as a function of (1-step/decay_steps)**power.  Once the final rate is reached, it remains there for
        the rest of optimization.

        Parameters
        ----------
        initial_rate: float
            the initial learning rate
        final_rate: float
            the final learning rate
        decay_steps: int
            the number of training steps over which the rate decreases from initial_rate to final_rate
        power: float
            the exponent controlling the shape of the decay
        """
        self.initial_rate = initial_rate
        self.final_rate = final_rate
        self.decay_steps = decay_steps
        self.power = power

    def _create_tf_tensor(self, global_step):
        import tensorflow as tf
        return tf.keras.optimizers.schedules.PolynomialDecay(
            initial_learning_rate=self.initial_rate,
            end_learning_rate=self.final_rate,
            decay_steps=self.decay_steps,
            power=self.power)(global_step)

    def _create_pytorch_schedule(self, optimizer):

        def f(step):
            t = min(step, self.decay_steps) / self.decay_steps
            return ((self.initial_rate - self.final_rate) *
                    (1 - t)**self.power) + self.final_rate

        import torch
        return torch.optim.lr_scheduler.LambdaLR(optimizer, f)

    def _create_jax_schedule(self):
        import optax
        return optax.polynomial_schedule(init_value=self.initial_rate,
                                         end_value=self.final_rate,
                                         power=self.power,
                                         transition_steps=self.decay_steps)


class LinearCosineDecay(LearningRateSchedule):
    """Applies linear cosine decay to the learning rate"""

    def __init__(self,
                 initial_rate: float,
                 decay_steps: int,
                 alpha: float = 0.0,
                 beta: float = 0.001,
                 num_periods: float = 0.5):
        """
        Parameters
        ----------
        learning_rate : float
        initial learning rate
        decay_steps : int
        number of steps to decay over
        num_periods : number of periods in the cosine part of the decay
        """

        self.initial_rate = initial_rate
        self.decay_steps = decay_steps
        self.alpha = alpha
        self.beta = beta
        self.num_periods = num_periods

    def _create_tf_tensor(self, global_step):
        import tensorflow as tf
        return tf.compat.v1.train.linear_cosine_decay(
            learning_rate=self.initial_rate,
            global_step=global_step,
            decay_steps=self.decay_steps,
            alpha=self.alpha,
            beta=self.beta,
            num_periods=self.num_periods)

    def _create_pytorch_schedule(self, optimizer):

        def f(step):
            t = min(step, self.decay_steps) / self.decay_steps
            linear_decay = 1 - t
            cosine_decay = 0.5 * (1 +
                                  math.cos(math.pi * 2 * self.num_periods * t))
            decayed = (self.alpha + linear_decay) * cosine_decay + self.beta
            return self.initial_rate * decayed

        import torch
        return torch.optim.lr_scheduler.LambdaLR(optimizer, f)

    def _create_jax_schedule(self):
        import optax
        return optax.cosine_decay_schedule(init_value=self.initial_rate,
                                           decay_steps=self.decay_steps,
                                           alpha=self.alpha)


class PiecewiseConstantSchedule(LearningRateSchedule):
    """Applies scheduler which multiplies by a constant factor on the boundaries"""

    def __init__(self,
                 initial_rate: float,
                 boundaries_and_scales: Optional[Dict[int, float]] = None):
        """
        Parameters
        ----------
        init_value : float
            initial learning rate
        boundaries_and_scales:
            A map from boundaries b_i to non-negative scaling factors f_i. For any step
            count s, the schedule returns init_v scaled by the product of all factors f_i
            such that b_i < s.
        """
        self.initial_rate = initial_rate
        self.boundaries_and_scales = boundaries_and_scales

    def _create_jax_schedule(self):
        import optax
        return optax.piecewise_constant_schedule(
            init_value=self.initial_rate,
            boundaries_and_scales=self.boundaries_and_scales)


class KFAC(Optimizer):
    """The Second order gradient optimiation algorithm which uses an approximation to calculate the inverse of the Fischer matrrix"""

    def __init__(self, **kwargs):
        """
        Parameters:
        -----------
        model: torch.nn.Module
            The model to be optimized.
        lr: float (default: 0.001)
            Learning rate for the optimizer.
        momentum: float (default: 0.9)
            Momentum for the optimizer.
        stat_decay: float (default: 0.95)
            Decay rate for the update of covariance matrix with mean.
        damping: float (default: 0.001)
            damping factor for the update of covariance matrix.
        kl_clip: float (default: 0.001)
            Clipping value for the update of covariance matrix.
        weight_decay: float (default: 0)
            weight decay for the optimizer.
        Tcov: int (default: 10)
            The number of steps to update the covariance matrix.
        Tinv: int (default: 100)
            The number of steps to calculate the inverse of covariance matrix.
        batch_averaged: bool (default: True)
            States whether to use batch averaged covariance matrix.
        mean: bool (default: False)
            States whether to use mean centered covariance matrix.
        """
        self.kwargs = kwargs

    def _create_pytorch_optimizer(self):
        from deepchem.models.torch_models.kfac_optimizer import KFACOptimizer
        if isinstance(self.learning_rate, LearningRateSchedule):
            self.kwargs['lr'] = self.learning_rate.initial_rate
        else:
            self.kwargs['lr'] = self.learning_rate
        return KFACOptimizer([self.kwargs])


class Lamb(Optimizer):
    """The Lamb optimization algorithm.

    It has been proposed in `Large Batch Optimization for Deep Learning:
    Training BERT in 76 minutes`__.
    """

    def __init__(self,
                 learning_rate: Union[float, LearningRateSchedule] = 0.001,
                 beta1: float = 0.9,
                 beta2: float = 0.999,
                 epsilon: float = 1e-08,
                 weight_decay: float = 0):
        """Construct an Adam optimizer.

        Parameters
        ----------
        learning_rate: float or LearningRateSchedule
            the learning rate to use for optimization
        beta1: float
            a parameter of the Lamb algorithm
        beta2: float
            a parameter of the Lamb algorithm
        epsilon: float
            a parameter of the Lamb algorithm
        weight_decay: float
            L2 penalty - a parameter of the Lamb algorithm

        Examples
        --------
        >>> import deepchem as dc
        >>> from deepchem.models import GCNModel
        >>> # preparing dataset
        >>> smiles = ["C1CCC1", "CCC"]
        >>> labels = [0., 1.]
        >>> featurizer = dc.feat.MolGraphConvFeaturizer()
        >>> X = featurizer.featurize(smiles)
        >>> dataset = dc.data.NumpyDataset(X=X, y=labels)
        >>> # training model
        >>> model = GCNModel(mode='classification', n_tasks=1,
        ...                  batch_size=16, learning_rate=0.001,
        ...                  optimizers=dc.models.optimizers.Lamb(learning_rate=0.01))
        >>> loss = model.fit(dataset, nb_epoch=5)

        References
        ----------
        Yang You and Jing Li and Sashank Reddi, et, al., "Large Batch Optimization for Deep Learning: Training BERT in 76 minutes", 2020,
        https://arxiv.org/abs/1904.00962
        """
        super(Lamb, self).__init__(learning_rate)
        self.beta1 = beta1
        self.beta2 = beta2
        self.epsilon = epsilon
        self.weight_decay = weight_decay

    def _create_pytorch_optimizer(self, params):
        if isinstance(self.learning_rate, LearningRateSchedule):
            lr = self.learning_rate.initial_rate
        else:
            lr = self.learning_rate
        return LambOptimizer(params,
                             lr=lr,
                             betas=(self.beta1, self.beta2),
                             eps=self.epsilon,
                             weight_decay=self.weight_decay)



import math
from typing import Optional, List, Dict, Tuple
import torch
import numpy as np


def get_basis(G,
              max_degree: int,
              compute_gradients: bool = False) -> Dict[str, torch.Tensor]:
    """
    Precompute the SE(3)-equivariant weight basis, \( W_J^{lk}(x) \).

    This function computes the equivariant weight basis used in SE(3)-equivariant
    convolutions, enabling the model to learn **rotation-equivariant filters.

    It is **called internally** by `get_basis_and_r()`, which computes **both**:
    - **Equivariant weight basis** \( W_J^{lk}(x) \).
    - **Inter-nodal distances** \( r_{ij} \) (used in attention & convolution).

    ---
    ### **Mathematical Background**
    This function follows Equation (8) from [SE(3)-Transformer paper](https://arxiv.org/pdf/2006.10503.pdf):

    \[
    W_J^{lk}(x) = Q_J \cdot Y_J(x)
    \]

    - \( Y_J(x) \): **Spherical Harmonic Basis**
    - \( Q_J \): **Basis transformation matrix**
    - \( J \): **Angular momentum index**
    - \( d_{in}, d_{out} \): **Feature degrees**

    This function precomputes basis functions for SE(3)-equivariant convolutions by calcutating spherical harmonic projections.

    Parameters
    ----------
    G:  dgl.DGLGraph
        DGL graph where `G.edata['d']` stores edge displacement vectors.
    max_degree: int
        Maximum degree (`l_max`) of equivariant tensors.
    compute_gradients: `bool`, optional, default=`False`
        If `True`: Enables gradient tracking for backpropagation.
        If `False`: Uses `torch.no_grad()` for efficiency.

    Returns
    -------
    basis: Dict[str, torch.Tensor]
        Dictionary mapping `'<d_in>,<d_out>'`: precomputed basis tensor.
        Shape: `(batch_size, 1, 2*d_out+1, 1, 2*d_in+1, num_bases)`

    Example
    -------

    >>> import torch
    >>> import dgl
    >>> from deepchem.utils.equivariance_utils import get_basis
    >>> from rdkit import Chem
    >>> import deepchem as dc
    >>> mol = Chem.MolFromSmiles('CCO')
    >>> featurizer = dc.feat.EquivariantGraphFeaturizer(fully_connected=True, embeded=True)
    >>> features = featurizer.featurize([mol])[0]
    >>> G = dgl.graph((features.edge_index[0], features.edge_index[1]))
    >>> G.ndata['f'] = torch.tensor(features.node_features, dtype=torch.float32).unsqueeze(-1)
    >>> G.ndata['x'] = torch.tensor(features.positions, dtype=torch.float32)
    >>> G.edata['d'] = torch.tensor(features.edge_features, dtype=torch.float32)
    >>> G.edata['w'] = torch.tensor(features.edge_weights, dtype=torch.float32)
    >>> basis = get_basis(G, max_degree=2)
    >>> print(basis.keys())
    dict_keys(['0,0', '0,1', '0,2', '1,0', '1,1', '1,2', '2,0', '2,1', '2,2'])
    """
    # Disable gradients unless explicitly enabled
    context = torch.enable_grad() if compute_gradients else torch.no_grad()

    with context:
        cloned_d = torch.clone(G.edata['d'])
        if G.edata['d'].requires_grad:
            cloned_d.requires_grad_()

        # Compute relative positional encodings in spherical coordinates
        r_ij = get_spherical_from_cartesian(cloned_d)

        # Compute Spherical Harmonics (Y_J) for all degrees up to 2*max_degree
        Y = precompute_sh(r_ij, 2 * max_degree)
        device = Y[0].device

        basis = {}
        for d_in in range(max_degree + 1):
            for d_out in range(max_degree + 1):
                K_Js = []
                for J in range(abs(d_in - d_out), d_in + d_out + 1):
                    # Compute basis transformation matrix Q_J
                    Q_J = basis_transformation_Q_J(J, d_in,
                                                   d_out).float().T.to(device)

                    # Apply projection using spherical harmonics
                    K_J = torch.matmul(Y[J], Q_J)
                    K_Js.append(K_J)

                # Reshape for broadcasting
                size = (-1, 1, 2 * d_out + 1, 1, 2 * d_in + 1,
                        2 * min(d_in, d_out) + 1)
                basis[f"{d_in},{d_out}"] = torch.stack(K_Js, -1).view(*size)

        return basis


def get_r(G) -> torch.Tensor:
    """
    Compute inter-nodal distances for a given DGL graph.

    This function computes the Euclidean distance between connected nodes
    based on edge feature `d`, which represents relative displacements.

    Parameters
    ----------
    G : dgl.DGLGraph
        The input graph where `G.edata['d']` contains edge displacement vectors.

    Returns
    -------
    torch.Tensor
        A tensor containing the computed inter-nodal distances for each edge,
        with shape `(num_edges, 1)`.

    Example
    -------
    >>> import torch
    >>> import dgl
    >>> from deepchem.utils.equivariance_utils import get_r
    >>> from rdkit import Chem
    >>> import deepchem as dc
    >>> mol = Chem.MolFromSmiles('CCO')
    >>> featurizer = dc.feat.EquivariantGraphFeaturizer(fully_connected=True, embeded=True)
    >>> features = featurizer.featurize([mol])[0]
    >>> G = dgl.graph((features.edge_index[0], features.edge_index[1]))
    >>> G.ndata['f'] = torch.tensor(features.node_features, dtype=torch.float32).unsqueeze(-1)
    >>> G.ndata['x'] = torch.tensor(features.positions, dtype=torch.float32)
    >>> G.edata['d'] = torch.tensor(features.edge_features, dtype=torch.float32)
    >>> G.edata['w'] = torch.tensor(features.edge_weights, dtype=torch.float32)
    >>> # Compute internodal distances
    >>> r = get_r(G)
    >>> print(r.shape)  # (num_edges, 1)
    torch.Size([6, 1])
    >>> print(r)
    tensor([[1.5317],
            [2.3267],
            [1.5317],
            [1.4096],
            [2.3267],
            [1.4096]])
    """
    cloned_d = torch.clone(G.edata['d'])
    if G.edata['d'].requires_grad:
        cloned_d.requires_grad_()
    return torch.sqrt(torch.sum(cloned_d**2, -1, keepdim=True))


def get_equivariant_basis_and_r(
    G,
    max_degree: int,
    compute_gradients: bool = False
) -> Tuple[Dict[str, torch.Tensor], torch.Tensor]:
    """
    Compute equivariant weight basis and inter-nodal distances for SE(3)-Transformers.

    This function computes:
    1. **Equivariant weight basis**: Required for SE(3)-equivariant convolutions.
    2. **Inter-nodal distances**: Used in attention mechanisms and basis function computation.

    This should be **called once per forward pass** to compute shared equivariant representations.

    Parameters
    ----------
    G : dgl.DGLGraph
        A DGL graph containing edge features.
    max_degree : int
        The maximum degree of the SE(3) tensor representation.
    compute_gradients : bool, optional (default=False)
        If True, enables gradient computation for the basis.

    Returns
    -------
    Tuple[Dict[str, torch.Tensor], torch.Tensor]
        - **basis (dict[str, torch.Tensor])**: Dictionary of equivariant bases,
          indexed by `'<d_in><d_out>'`, where `d_in` and `d_out` are feature degrees.
        - **r (torch.Tensor)**: Inter-nodal distance tensor with shape `(num_edges, 1)`.

    Example
    -------
    >>> import torch
    >>> import dgl
    >>> from deepchem.utils.equivariance_utils import get_equivariant_basis_and_r
    >>> from rdkit import Chem
    >>> import deepchem as dc
    >>> mol = Chem.MolFromSmiles('CCO')
    >>> featurizer = dc.feat.EquivariantGraphFeaturizer(fully_connected=True, embeded=True)
    >>> features = featurizer.featurize([mol])[0]
    >>> G = dgl.graph((features.edge_index[0], features.edge_index[1]))
    >>> G.ndata['f'] = torch.tensor(features.node_features, dtype=torch.float32).unsqueeze(-1)
    >>> G.ndata['x'] = torch.tensor(features.positions, dtype=torch.float32)
    >>> G.edata['d'] = torch.tensor(features.edge_features, dtype=torch.float32)
    >>> G.edata['w'] = torch.tensor(features.edge_weights, dtype=torch.float32)
    >>> # Compute basis and distances
    >>> basis, r = get_equivariant_basis_and_r(G, max_degree=2)
    >>> print(r.shape)  # Expected: (num_edges, 1)
    torch.Size([6, 1])
    >>> print(basis.keys())  # Expected: dict
    dict_keys(['0,0', '0,1', '0,2', '1,0', '1,1', '1,2', '2,0', '2,1', '2,2'])
    """
    basis = get_basis(G, max_degree, compute_gradients)
    r = get_r(G)
    return basis, r


def semifactorial(x: int) -> float:
    """
    Compute the semifactorial function x!!, defined as:
    x!! = x * (x-2) * (x-4) * ...
    Parameters
    ----------
    x : int
        A positive integer.
    Returns
    -------
    float
        The value of x!!, computed iteratively.
    Examples
    --------
    >>> semifactorial(5)
    15.0
    >>> semifactorial(6)
    48.0
    """
    # edge cases
    if x == 0 or x == 1:
        return 1.0

    y = 1.0
    for n in range(x, 1, -2):
        y *= n
    return y


def pochhammer(x: int, k: int) -> float:
    """
    Compute the Pochhammer symbol (x)_k , defined as:
    (x)_k = x * (x+1) * (x+2) * ... * (x+k-1).
    Parameters
    ----------
    x : int
        The starting integer of the sequence.
    k : int
        The number of terms in the product.
    Returns
    -------
    float
        The Pochhammer symbol value.
    Examples
    --------
    >>> pochhammer(3, 4)
    360.0
    >>> pochhammer(5, 2)
    30.0
    """
    # handle edge case (k=0)
    if k == 0:
        return 1.0

    xf = float(x)
    for n in range(x + 1, x + k):
        xf *= n
    return xf


def lpmv(d: int, m: int, x: torch.Tensor) -> torch.Tensor:
    """
    Compute the associated Legendre function P_l^m(x), including
    the Condon-Shortley phase. The implementation includes the base case
    for P_m^m(x), P_{m+1}^m(x), and uses a recurrence relation
    for higher degrees.
    Parameters
    ----------
    l : int
        Degree of the Legendre polynomial.
    m : int
        Order of the Legendre polynomial.
    x : torch.Tensor
        A tensor of shape (N,) representing the input values for x,
        where ( -1 <= x <= 1).
    Returns
    -------
    torch.Tensor
        A tensor of shape (N,) containing the values of P_l^m(x).
    Examples
    --------
    >>> lpmv(2, 1, torch.tensor([0.5]))
    tensor([-1.2990])
    """
    m_abs = abs(m)
    if m_abs > d:
        return torch.zeros_like(x)

    # base case: P_m^m(x)
    p_mm = ((-1)**m_abs * semifactorial(2 * m_abs - 1)) * torch.pow(
        1 - x * x, m_abs / 2)

    if d == m_abs:
        return p_mm

    # Compute P_{m+1}^m(x)
    p_m1m = x * (2 * m_abs + 1) * p_mm

    if d == m_abs + 1:
        return p_m1m

    # Recurrence relation for higher degrees (l)
    p_lm = p_m1m.clone()
    for i in range(m_abs + 2, d + 1):
        p_lm = ((2 * i - 1) * x * p_m1m - (i + m_abs - 1) * p_mm) / (i - m_abs)
        p_mm, p_m1m = p_m1m, p_lm

    # Negative values of m
    if m < 0:
        p_lm *= ((-1)**m) * pochhammer(d - m + 1, -2 * m)

    return p_lm


class SphericalHarmonics:
    """
    A class for computing real tesseral spherical harmonics, including
    the Condon-Shortley phase.
    This implementation is inspired by the SE(3)-Transformer library by Fabian Fuchs,
    which provides efficient computation of spherical harmonics and related transformations.
    For more details, see the SE(3)-Transformer repository:
    https://github.com/FabianFuchsML/se3-transformer-public

    Methods
    -------
    get_element(l, m, theta, phi)
        Compute the tesseral spherical harmonic Y_l^m(theta, phi).
    get(l, theta, phi)
        Compute all spherical harmonics of degree l for given angles.
    """

    def __init__(self):
        self.leg = {}

    def clear(self) -> None:
        """Clear cached Legendre polynomial values to save RAM memory."""
        self.leg = {}

    def get_element(self, d: int, m: int, theta: torch.Tensor,
                    phi: torch.Tensor) -> torch.Tensor:
        """
        Compute a single tesseral spherical harmonic Y_l^m(theta, phi).
        Parameters
        ----------
        l : int
            Degree of the spherical harmonic.
        m : int
            Order of the spherical harmonic.
        theta : torch.Tensor
            Tensor of polar angles (collatitude) in radians.
        phi : torch.Tensor
            Tensor of azimuthal angles (longitude) in radians.
        Returns
        -------
        torch.Tensor
            Tensor of the spherical harmonic values, same shape as `theta`.
        Examples
        --------
        >>> theta = torch.tensor([0.0, math.pi / 2])
        >>> phi = torch.tensor([0.0, math.pi])
        >>> SphericalHarmonics().get_element(1, 0, theta, phi)
        tensor([ 4.8860e-01, -2.1357e-08])
        """

        N = math.sqrt((2 * d + 1) / (4 * math.pi))
        leg = lpmv(d, abs(m), torch.cos(theta))
        if m == 0:
            return N * leg
        elif m > 0:
            Y = torch.cos(m * phi) * leg
        else:
            Y = torch.sin(abs(m) * phi) * leg
        N *= math.sqrt(2.0 / pochhammer(d - abs(m) + 1, 2 * abs(m)))
        Y *= N
        return Y

    def get(self,
            d: int,
            theta: torch.Tensor,
            phi: torch.Tensor,
            refresh=True) -> torch.Tensor:
        """
        Compute all spherical harmonics of degree l.
        Parameters
        ----------
        l : int
            Degree of the spherical harmonics.
        theta : torch.Tensor
            Tensor of polar angles (collatitude) in radians.
        phi : torch.Tensor
            Tensor of azimuthal angles (longitude) in radians.
        Returns
        -------
        torch.Tensor
            A tensor of shape [*theta.shape, 2 * l + 1].
        Examples
        --------
        >>> theta = torch.tensor([0.0, math.pi / 2])
        >>> phi = torch.tensor([0.0, math.pi])
        >>> SphericalHarmonics().get(1, theta, phi)
        tensor([[-0.0000e+00,  4.8860e-01, -0.0000e+00],
                [ 4.2715e-08, -2.1357e-08,  4.8860e-01]])
        """
        if refresh:
            self.clear()
        results = []
        for m in range(-d, d + 1):
            results.append(self.get_element(d, m, theta, phi))
        return torch.stack(results, dim=-1)


def irr_repr(order: int,
             alpha: float,
             beta: float,
             gamma: float,
             dtype: Optional[torch.dtype] = None) -> torch.Tensor:
    """
    Compute the irreducible representation of the special orthogonal group SO(3).
    This function computes the Wigner D-matrix for the given order and angles (alpha, beta, gamma).
    It is compatible with composition and spherical harmonics computations.
    The output is a 2D tensor of shape `(2*order + 1, 2*order + 1)` where each element represents
    the transformation coefficients of the Wigner D-matrix. The values correspond to the rotational
    components for the given order and Euler angles.
    Parameters
    ----------
    order : int
        The order of the representation.
    alpha : float
        The first Euler angle (rotation about the Y axis).
    beta : float
        The second Euler angle (rotation about the X axis).
    gamma : float
        The third Euler angle (rotation about the Y axis).
    dtype : torch.dtype, optional
        The desired data type of the resulting tensor. If None, the default dtype is used.
    Returns
    -------
    torch.Tensor
        The irreducible representation matrix of SO(3) for the specified order and angles.
    Examples
    --------
    >>> irr_repr(1, 0.1, 0.2, 0.3)
    tensor([[ 0.9216,  0.0587,  0.3836],
            [ 0.0198,  0.9801, -0.1977],
            [-0.3875,  0.1898,  0.9021]])
    """
    result = wigner_D(order, torch.tensor(alpha), torch.tensor(beta),
                      torch.tensor(gamma))[0]
    return result.clone().detach() if dtype is None else result.clone().detach(
    ).to(dtype)


def su2_generators(k: int) -> torch.Tensor:
    """Generate the generators of the special unitary group SU(2) in a given representation.

    The function computes the generators of the SU(2) group for a specific representation
    determined by the value of 'k'. These generators are commonly used in the study of
    quantum mechanics, angular momentum, and related areas of physics and mathematics.
    The generators are represented as matrices.

    The SU(2) group is a fundamental concept in quantum mechanics and symmetry theory.
    The generators of the group, denoted as J_x, J_y, and J_z, represent the three
    components of angular momentum operators. These generators play a key role in
    describing the transformation properties of physical systems under rotations.

    The returned tensor contains three matrices corresponding to the x, y, and z generators,
    usually denoted as J_x, J_y, and J_z. These matrices form a basis for the Lie algebra
    of the SU(2) group.

    In linear algebra, specifically within the context of quantum mechanics, lowering and
    raising operators are fundamental concepts that play a crucial role in altering the
    eigenvalues of certain operators while acting on quantum states. These operators are
    often referred to collectively as "ladder operators."

    A lowering operator is an operator that, when applied to a quantum state, reduces the
    eigenvalue associated with a particular observable. In the context of SU(2), the lowering
    operator corresponds to J_-.

    Conversely, a raising operator is an operator that increases the eigenvalue of an
    observable when applied to a quantum state. In the context of SU(2), the raising operator
    corresponds to J_+.

    The z-generator matrix represents the component of angular momentum along the z-axis,
    often denoted as J_z. It commutes with both J_x and J_y and is responsible for quantizing
    the angular momentum.

    Note that the dimensions of the returned tensor will be (3, 2j+1, 2j+1), where each matrix
    has a size of (2j+1) x (2j+1).
    Parameters
    ----------
    k : int
        The representation index, which determines the order of the representation.

    Returns
    -------
    torch.Tensor
        A stack of three SU(2) generators, corresponding to J_x, J_z, and J_y.

    Notes
    -----
    A generating set of a group is a subset $S$ of the group $G$ such that every element
    of $G$ can be expressed as a combination (under the group operation) of finitely many
    elements of the subset $S$ and their inverses.

    The special unitary group $SU_n(q)$ is the set of $n*n$ unitary matrices with determinant
    +1. $SU(2)$ is homeomorphic with the orthogonal group $O_3^+(2)$. It is also called the
    unitary unimodular group and is a Lie group.

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Ladder_operator
    .. [2] https://en.wikipedia.org/wiki/Special_unitary_group#The_group_SU(2)
    .. [3] https://en.wikipedia.org/wiki/Generating_set_of_a_group
    .. [4] https://mathworld.wolfram.com/SpecialUnitaryGroup

    Examples
    --------
    >>> su2_generators(1)
    tensor([[[ 0.0000+0.0000j,  0.7071+0.0000j,  0.0000+0.0000j],
             [-0.7071+0.0000j,  0.0000+0.0000j,  0.7071+0.0000j],
             [ 0.0000+0.0000j, -0.7071+0.0000j,  0.0000+0.0000j]],
    <BLANKLINE>
            [[-0.0000-1.0000j,  0.0000+0.0000j,  0.0000+0.0000j],
             [ 0.0000+0.0000j,  0.0000+0.0000j,  0.0000+0.0000j],
             [ 0.0000+0.0000j,  0.0000+0.0000j,  0.0000+1.0000j]],
    <BLANKLINE>
            [[ 0.0000-0.0000j,  0.0000+0.7071j,  0.0000-0.0000j],
             [ 0.0000+0.7071j,  0.0000-0.0000j,  0.0000+0.7071j],
             [ 0.0000-0.0000j,  0.0000+0.7071j,  0.0000-0.0000j]]])
    """
    # Generate the raising operator matrix
    m = torch.arange(-k, k)
    raising = torch.diag(-torch.sqrt(k * (k + 1) - m * (m + 1)), diagonal=-1)

    # Generate the lowering operator matrix
    m = torch.arange(-k + 1, k + 1)
    lowering = torch.diag(torch.sqrt(k * (k + 1) - m * (m - 1)), diagonal=1)

    # Generate the z-generator matrix
    m = torch.arange(-k, k + 1)
    z_generator = torch.diag(1j * m)

    # Combine the matrices to form the x, z, and y generators
    x_generator = 0.5 * (raising + lowering)  # x (usually)
    y_generator = -0.5j * (raising - lowering)  # -y (usually)

    # Stack the generators along the first dimension to create a tensor
    generators = torch.stack([x_generator, z_generator, y_generator], dim=0)

    return generators


def get_matrix_kernel(A: torch.Tensor, eps: float = 1e-10) -> torch.Tensor:
    """
    Compute an orthonormal basis of the kernel (x_1, x_2, ...).

    This function calculates the null space (kernel) of the given matrix A, such that:
        A x_i = 0
        scalar_product(x_i, x_j) = delta_ij

    Parameters
    ----------
    A : torch.Tensor
        The input matrix.
    eps : float, optional
        Tolerance for singular values considered as zero (default is 1e-10).

    Returns
    -------
    torch.Tensor
        A matrix where each row is a basis vector of the kernel of A.

    Examples
    --------
    >>> from deepchem.utils.equivariance_utils import get_matrix_kernel
    >>> A = torch.tensor([[1.0, 2.0, 3.0],
    ...                   [2.0, 4.0, 6.0],
    ...                   [3.0, 6.0, 9.0]])
    >>> get_matrix_kernel(A)
    tensor([[ 0.0000, -0.8321,  0.5547],
            [ 0.9636, -0.1482, -0.2224]])
    """
    _, s, v = torch.svd(A)

    kernel = v.t()[s < eps]
    return kernel


def get_matrices_kernel(As: List[torch.Tensor],
                        eps: float = 1e-10) -> torch.Tensor:
    """
    Compute the common kernel of all the input matrices.

    This function computes the shared null space of a collection of matrices.

    Parameters
    ----------
    As : List[torch.Tensor]
        List of input matrices.
    eps : float, optional
        Tolerance for singular values considered as zero (default is 1e-10).

    Returns
    -------
    torch.Tensor
        A matrix where each row is a basis vector of the common kernel.
    """
    return get_matrix_kernel(torch.cat(As, dim=0), eps)


def basis_transformation_Q_J(J: int,
                             order_in: int,
                             order_out: int,
                             eps: float = 1e-10,
                             num_samples: int = 5,
                             random_angle_higher: float = 6.2,
                             random_angle_lower: float = 0.2) -> torch.Tensor:
    """
    Compute one part of the Q^-1 matrix for the article.

    This function computes the spherical harmonics projection matrix for the
    Sylvester equation in the subspace J needed in for the weight basis in SE(3)-Transformer model.

    References:
    -----------
    - SE(3)-Transformers: 3D Roto-Translation Equivariant Attention Networks
    Fabian B. Fuchs, Daniel E. Worrall, Volker Fischer, Max Welling
    NeurIPS 2020, https://arxiv.org/abs/2006.10503

    Parameters
    ----------
    J : int
        Order of the spherical harmonics.
    order_in : int
        Order of the input representation.
    order_out : int
        Order of the output representation.
    version : int, optional
        Version of the computation (default is 3).
    eps : float, optional
        Tolerance for singular values considered as zero (default is 1e-10).
    num_samples : int, optional
        Number of samples to generate for random angles (default is 5).
    random_angle_higher : float, optional
        Upper limit for generating random angles (default is 6.2).
    random_angle_lower : float, optional
        Lower limit for generating random angles (default is 0.2).

    Returns
    -------
    torch.Tensor
        A tensor of shape [(m_out * m_in), m], where m = 2 * J + 1.

    Examples
    --------
    >>> from deepchem.utils.equivariance_utils import basis_transformation_Q_J
    >>> basis_transformation_Q_J(1, 1, 1).shape
    torch.Size([9, 3])
    """
    original_dtype = torch.get_default_dtype()
    torch.set_default_dtype(torch.float64)

    def _R_tensor(a: float, b: float, c: float) -> torch.Tensor:
        """
        Compute the Kronecker product of irreducible representations.

        This function calculates the Kronecker product of two irreducible
        representations (input and output orders) for a given set of rotation angles.

        Parameters
        ----------
        a : float
            Rotation angle around the x-axis.
        b : float
            Rotation angle around the y-axis.
        c : float
            Rotation angle around the z-axis.

        Returns
        -------
        torch.Tensor
            The Kronecker product of the irreducible representations of the
            given rotation angles.

        Examples
        --------
        >>> _R_tensor(1.0, 2.0, 3.0).shape
        torch.Size([...])
        return kron(irr_repr(order_out, a, b, c), irr_repr(order_in, a, b, c))
        """
        return kron(irr_repr(order_out, a, b, c), irr_repr(order_in, a, b, c))

    def _sylvester_submatrix(J: int, a: float, b: float,
                             c: float) -> torch.Tensor:
        """
        Generate the Kronecker product matrix for solving the Sylvester equation in subspace J.

        This function constructs a Kronecker product matrix that helps solve
        the Sylvester equation for the given subspace J. The equation ensures
        the transformation is valid within the subspace.

        Parameters
        ----------
        J : int
            Order of the spherical harmonics.
        a : float
            Rotation angle around the x-axis.
        b : float
            Rotation angle around the y-axis.
        c : float
            Rotation angle around the z-axis.

        Returns
        -------
        torch.Tensor
            A rank-deficient matrix for use in solving the Sylvester equation.

        Examples
        --------
        >>> _sylvester_submatrix(1, 1.0, 2.0, 3.0).shape
        torch.Size([...])
        """
        R_tensor = _R_tensor(a, b, c)
        R_irrep_J = irr_repr(J, a, b, c)
        return kron(R_tensor, torch.eye(R_irrep_J.size(0))) - \
                   kron(torch.eye(R_tensor.size(0)), R_irrep_J.t())

    random_angles = np.random.uniform(random_angle_lower,
                                      random_angle_higher,
                                      size=(num_samples, 3))

    null_space = get_matrices_kernel(
        [_sylvester_submatrix(J, a, b, c) for a, b, c in random_angles])
    Q_J = null_space[0]
    Q_J = Q_J.view((2 * order_out + 1) * (2 * order_in + 1), 2 * J + 1)
    torch.set_default_dtype(original_dtype)

    return Q_J


def get_spherical_from_cartesian(cartesian: torch.Tensor,
                                 divide_radius_by: float = 1.0) -> torch.Tensor:
    """
    Convert Cartesian coordinates to spherical coordinates.

    Parameters
    ----------
    cartesian : torch.Tensor
        Cartesian coordinates tensor of shape [..., 3].
    divide_radius_by : float, optional
        Factor by which to divide the radius (default is 1.0).

    Returns
    -------
    torch.Tensor
        Spherical coordinates tensor of shape [..., 3] with [radius, azimuth (phi), elevation (theta)].

    Examples
    --------
    >>> from deepchem.utils.equivariance_utils import get_spherical_from_cartesian
    >>> cartesian = torch.tensor([[1.0, 1.0, 1.0]])
    >>> get_spherical_from_cartesian(cartesian)
    tensor([[1.7321, 0.7854, 0.9553]])

    >>> cartesian = torch.tensor([[0.0, 0.0, 1.0]])  # Point on Z-axis
    >>> get_spherical_from_cartesian(cartesian)
    tensor([[1.0000, 0.0000, 1.5708]])

    >>> cartesian = torch.tensor([[0.0, 0.0, -1.0]])  # Point on negative Z-axis
    >>> get_spherical_from_cartesian(cartesian)
    tensor([[1.0000, 3.1416, 1.5708]])
    """
    spherical = torch.zeros_like(cartesian)

    ind_radius = 0
    ind_alpha = 1
    ind_beta = 2

    cartesian_x = 2
    cartesian_y = 0
    cartesian_z = 1

    r_xy = cartesian[..., cartesian_x]**2 + cartesian[..., cartesian_y]**2

    spherical[..., ind_beta] = torch.atan2(torch.sqrt(r_xy),
                                           cartesian[..., cartesian_z])
    spherical[..., ind_alpha] = torch.atan2(cartesian[..., cartesian_y],
                                            cartesian[..., cartesian_x])

    if divide_radius_by == 1.0:
        spherical[..., ind_radius] = torch.sqrt(r_xy +
                                                cartesian[..., cartesian_z]**2)
    else:
        spherical[..., ind_radius] = torch.sqrt(
            r_xy + cartesian[..., cartesian_z]**2) / divide_radius_by

    return spherical


def kron(a: torch.Tensor, b: torch.Tensor) -> torch.Tensor:
    """
    Computes the Kronecker product of two tensors needed to comput Q_J matrix.

    References:
    -----------
    - SE(3)-Transformers: 3D Roto-Translation Equivariant Attention Networks
    Fabian B. Fuchs, Daniel E. Worrall, Volker Fischer, Max Welling
    NeurIPS 2020, https://arxiv.org/abs/2006.10503

    Parameters
    ----------
    a : torch.Tensor
        First input tensor of shape [*, m, n].
    b : torch.Tensor
        Second input tensor of shape [*, p, q].

    Returns
    -------
    torch.Tensor
        The Kronecker product of `a` and `b`.

    Examples
    --------
    >>> from deepchem.utils.equivariance_utils import kron
    >>> A = torch.tensor([[1, 2], [3, 4]])
    >>> B = torch.tensor([[0, 5], [6, 7]])
    >>> kron(A, B)
    tensor([[ 0,  5,  0, 10],
            [ 6,  7, 12, 14],
            [ 0, 15,  0, 20],
            [18, 21, 24, 28]])
    """
    if a.ndimension() == 2 and b.ndimension() == 2:
        return torch.einsum("ij,kl->ikjl", (a, b)).reshape(
            a.size(0) * b.size(0),
            a.size(1) * b.size(1))

    siz1 = torch.Size(torch.tensor(a.shape[-2:]) * torch.tensor(b.shape[-2:]))
    res = a.unsqueeze(-1).unsqueeze(-3) * b.unsqueeze(-2).unsqueeze(-4)
    siz0 = res.shape[:-4]
    return res.reshape(siz0 + siz1)


def precompute_sh(r_ij: torch.Tensor, max_J: int) -> dict:
    """
    Precompute spherical harmonics up to a given order used in the forward pass of
    SE(3)-Transformer model.

    References:
    -----------
    - SE(3)-Transformers: 3D Roto-Translation Equivariant Attention Networks
    Fabian B. Fuchs, Daniel E. Worrall, Volker Fischer, Max Welling
    NeurIPS 2020, https://arxiv.org/abs/2006.10503

    Parameters
    ----------
    r_ij : torch.Tensor
        Relative positions tensor.
    max_J : int
        Maximum order of the spherical harmonics.

    Returns
    -------
    dict
        A dictionary where each key corresponds to an order J and the value is a tensor
        of shape [B, N, K, 2J+1].

    Examples
    --------
    >>> from deepchem.utils.equivariance_utils import precompute_sh
    >>> r_ij = torch.tensor([[1.0, 0.5, 1.0]])  # Example spherical coordinates (radius, phi, theta)
    >>> precompute_sh(r_ij, max_J=2)
    {0: tensor([[0.2821]]), 1: tensor([[-0.1971, -0.2640, -0.3608]]), 2: tensor([[ 0.3255,  0.2381, -0.0392,  0.4359,  0.2090]])}
    """
    i_alpha = 1
    i_beta = 2

    Y_Js = {}
    sh = SphericalHarmonics()

    for J in range(max_J + 1):
        Y_Js[J] = sh.get(J,
                         theta=math.pi - r_ij[..., i_beta],
                         phi=r_ij[..., i_alpha],
                         refresh=False)

    sh.clear()
    return Y_Js


def change_basis_real_to_complex(
        k: int,
        dtype: Optional[torch.dtype] = None,
        device: Optional[torch.device] = None) -> torch.Tensor:
    r"""Construct a transformation matrix to change the basis from real to complex spherical harmonics.

    This function constructs a transformation matrix Q that converts real spherical
    harmonics into complex spherical harmonics.
    It operates on the basis functions $Y_{\ell m}$ and $Y_{\ell}^{m}$, and accounts
    for the relationship between the real and complex forms of these harmonics
    as defined in the provided mathematical expressions.

    The resulting transformation matrix Q is used to change the basis of vectors or tensors of real spherical harmonics to
    their complex counterparts.

    Parameters
    ----------
    k : int
        The representation index, which determines the order of the representation.
    dtype : torch.dtype, optional
        The data type for the output tensor. If not provided, the
        function will infer it. Default is None.
    device : torch.device, optional
        The device where the output tensor will be placed. If not provided,
        the function will use the default device. Default is None.

    Returns
    -------
    torch.Tensor
        A transformation matrix Q that changes the basis from real to complex spherical harmonics.

    Notes
    -----
    Spherical harmonics Y_l^m are a family of functions that are defined on the surface of a
    unit sphere. They are used to represent various physical and mathematical phenomena that
    exhibit spherical symmetry. The indices l and m represent the degree and order of the
    spherical harmonics, respectively.

    The conversion from real to complex spherical harmonics is achieved by applying specific
    transformation coefficients to the real-valued harmonics. These coefficients are derived
    from the properties of spherical harmonics.

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Spherical_harmonics#Real_form

    Examples
    --------
    # The transformation matrix generated is used to change the basis of a vector of
    # real spherical harmonics with representation index 1 to complex spherical harmonics.
    >>> change_basis_real_to_complex(1)
    tensor([[-0.7071+0.0000j,  0.0000+0.0000j,  0.0000-0.7071j],
            [ 0.0000+0.0000j,  0.0000-1.0000j,  0.0000+0.0000j],
            [-0.7071+0.0000j,  0.0000+0.0000j,  0.0000+0.7071j]])
    """
    q = torch.zeros((2 * k + 1, 2 * k + 1), dtype=torch.complex128)

    # Construct the transformation matrix Q for m in range(-k, 0)
    for m in range(-k, 0):
        q[k + m, k + abs(m)] = 1 / 2**0.5
        q[k + m, k - abs(m)] = complex(-1j / 2**0.5)  # type: ignore

    # Set the diagonal elements for m = 0
    q[k, k] = 1

    # Construct the transformation matrix Q for m in range(1, k + 1)
    for m in range(1, k + 1):
        q[k + m, k + abs(m)] = (-1)**m / 2**0.5
        q[k + m, k - abs(m)] = complex(1j * (-1)**m / 2**0.5)  # type: ignore

    # Apply the factor of (-1j)**k to make the Clebsch-Gordan coefficients real
    q = (-1j)**k * q

    # Handle dtype and device options
    if dtype is None:
        default_type = torch.empty(0).dtype
        if default_type == torch.float32:
            dtype = torch.complex64
        elif default_type == torch.float64:
            dtype = torch.complex128
    if device is None:
        device = torch.empty(0).device

    # Ensure the tensor is contiguous and on the specified device
    return q.to(
        dtype=dtype,
        device=device,
        copy=True,
        memory_format=torch.contiguous_format)  # type: ignore[call-overload]


def so3_generators(k: int) -> torch.Tensor:
    """Construct the generators of the SO(3) Lie algebra for a given quantum angular momentum.

    The function generates the generators of the special orthogonal group SO(3), which represents the group
    of rotations in three-dimensional space. Its Lie algebra, which consists of the generators of
    infinitesimal rotations, is often used in physics to describe angular momentum operators.
    The generators of the Lie algebra can be related to the SU(2) group, and this function uses
    a transformation to convert the SU(2) generators to the SO(3) basis.

    The primary significance of the SO(3) group lies in its representation of three-dimensional
    rotations. Each matrix in SO(3) corresponds to a unique rotation, capturing the intricate
    ways in which objects can be oriented in 3D space. This concept finds application in
    numerous fields, ranging from physics to engineering.

    Parameters
    ----------
     k : int
        The representation index, which determines the order of the representation.

    Returns
    -------
    torch.Tensor
        A stack of three SO(3) generators, corresponding to J_x, J_z, and J_y.

    Notes
    -----
    The special orthogonal group $SO_n(q)$ is the subgroup of the elements of general orthogonal
    group $GO_n(q)$ with determinant 1. $SO_3$ (often written $SO(3)$) is the rotation group
    for three-dimensional space.

    These matrices are orthogonal, which means their rows and columns form mutually perpendicular
    unit vectors. This preservation of angles and lengths makes orthogonal matrices fundamental
    in various mathematical and practical applications.

    The "special" part of $SO(3)$ refers to the determinant of these matrices being $+1$. The
    determinant is a scalar value that indicates how much a matrix scales volumes.
    A determinant of $+1$ ensures that the matrix represents a rotation in three-dimensional
    space without involving any reflection or scaling operations that would reverse the orientation of space.

    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Special_orthogonal_group
    .. [2] https://en.wikipedia.org/wiki/3D_rotation_group#Connection_between_SO(3)_and_SU(2)
    .. [3] https://www.pas.rochester.edu/assets/pdf/undergraduate/su-2s_double_covering_of_so-3.pdf

    Examples
    --------
    >>> so3_generators(1)
    tensor([[[ 0.0000,  0.0000,  0.0000],
             [ 0.0000,  0.0000, -1.0000],
             [ 0.0000,  1.0000,  0.0000]],
    <BLANKLINE>
            [[ 0.0000,  0.0000,  1.0000],
             [ 0.0000,  0.0000,  0.0000],
             [-1.0000,  0.0000,  0.0000]],
    <BLANKLINE>
            [[ 0.0000, -1.0000,  0.0000],
             [ 1.0000,  0.0000,  0.0000],
             [ 0.0000,  0.0000,  0.0000]]])
    """
    # Get the SU(2) generators for the given quantum angular momentum (spin) value.
    X = su2_generators(k)

    # Get the transformation matrix to change the basis from real to complex spherical harmonics.
    Q = change_basis_real_to_complex(k)

    # Convert the SU(2) generators to the SO(3) basis using the transformation matrix Q.
    # X represents the SU(2) generators, and Q is the transformation matrix from real to complex spherical harmonics.
    # The resulting X matrix will be the SO(3) generators in the complex basis.
    X = torch.conj(Q.T) @ X @ Q

    # Return the real part of the SO(3) generators to ensure they are purely real.
    return torch.real(X)


def wigner_D(k: int, alpha: torch.Tensor, beta: torch.Tensor,
             gamma: torch.Tensor) -> torch.Tensor:
    """Wigner D matrix representation of the SO(3) rotation group.

    The function computes the Wigner D matrix representation of the SO(3) rotation group
    for a given representation index 'k' and rotation angles 'alpha', 'beta', and 'gamma'.
    The resulting matrix satisfies properties of the SO(3) group representation.

    Parameters
    ----------
    k : int
        The representation index, which determines the order of the representation.
    alpha : torch.Tensor
        Rotation angles (in radians) around the Y axis, applied third.
    beta : torch.Tensor
        Rotation angles (in radians) around the X axis, applied second.
    gamma : torch.Tensor)
        Rotation angles (in radians) around the Y axis, applied first.

    Returns
    -------
    torch.Tensor
        The Wigner D matrix of shape (#angles, 2k+1, 2k+1).

    Notes
    -----
    The Wigner D-matrix is a unitary matrix in an irreducible representation
    of the groups SU(2) and SO(3).

    The Wigner D-matrix is used in quantum mechanics to describe the action
    of rotations on states of particles with angular momentum. It is a key
    concept in the representation theory of the rotation group SO(3), and
    it plays a crucial role in various physical contexts.

    Examples
    --------
    >>> k = 1
    >>> alpha = torch.tensor([0.1, 0.2])
    >>> beta = torch.tensor([0.3, 0.4])
    >>> gamma = torch.tensor([0.5, 0.6])
    >>> wigner_D_matrix = wigner_D(k, alpha, beta, gamma)
    >>> wigner_D_matrix
    tensor([[[ 0.8275,  0.1417,  0.5433],
             [ 0.0295,  0.9553, -0.2940],
             [-0.5607,  0.2593,  0.7863]],
    <BLANKLINE>
            [[ 0.7056,  0.2199,  0.6737],
             [ 0.0774,  0.9211, -0.3817],
             [-0.7044,  0.3214,  0.6329]]])
    """
    # Ensure that alpha, beta, and gamma have the same shape for broadcasting.
    alpha, beta, gamma = torch.broadcast_tensors(alpha, beta, gamma)

    # Ensure the angles are within the range [0, 2*pi) using modulo.
    alpha = alpha[..., None, None] % (2 * math.pi)
    beta = beta[..., None, None] % (2 * math.pi)
    gamma = gamma[..., None, None] % (2 * math.pi)

    # Get the SO(3) generators for the given quantum angular momentum (spin) value 'k'.
    X = so3_generators(k)

    # Calculate the Wigner D matrix using the matrix exponential of the generators
    # and the rotation angles alpha, beta, and gamma in the appropriate order.
    D_matrix = torch.matrix_exp(gamma * (X[1].unsqueeze(0))) @ torch.matrix_exp(
        beta * (X[0].unsqueeze(0))) @ torch.matrix_exp(alpha *
                                                       (X[1].unsqueeze(0)))
    return D_matrix


def commutator(A: torch.Tensor, B: torch.Tensor) -> torch.Tensor:
    """Compute the commutator of two matrices.

    Parameters
    ----------
    A : torch.Tensor
        The first matrix.
    B : torch.Tensor
        The second matrix.

    Returns
    -------
    torch.Tensor
        The commutator of the two matrices.

    Examples
    --------
    >>> A = torch.tensor([[1, 2], [3, 4]])
    >>> B = torch.tensor([[5, 6], [7, 8]])
    >>> commutator(A, B)
    tensor([[ -4, -12],
            [ 12,   4]])
    """
    return torch.matmul(A, B) - torch.matmul(B, A)



import torch
from typing import List
from deepchem.utils.differentiation_utils import EditableModule, LinearOperator, symeig


class OrbitalOrthogonalizer(EditableModule):
    """Convert orbital to another type of orbital by orthogonalizing the basis sets.

    Examples
    --------
    >>> import torch
    >>> from deepchem.utils.dft_utils import OrbitalOrthogonalizer
    >>> ovlp = torch.tensor([[1.0, 0.5], [0.5, 1.0]])
    >>> orthozer = OrbitalOrthogonalizer(ovlp)
    >>> orthozer.nao()
    2
    >>> mat = torch.tensor([[1.0, 0.5], [0.5, 1.0]])
    >>> orthozer.convert2(mat)
    tensor([[1.0000, 0.0000],
            [0.0000, 1.0000]])

    """

    def __init__(self, ovlp: torch.Tensor, threshold: float = 1e-6):
        """Initialize the orbital orthogonalizer.

        Parameters
        ----------
        ovlp: torch.Tensor
            Overlap matrix of the original orbital basis sets.
        threshold: float
            Threshold to determine the accuracy of the overlap matrix.
        """
        ovlp_eival, ovlp_eivec = symeig(
            LinearOperator.m(ovlp, is_hermitian=True))
        acc_idx = ovlp_eival > threshold
        orthozer = ovlp_eivec[..., acc_idx] * (ovlp_eival[acc_idx])**(
            -0.5)  # (nao, nao2)
        self._orthozer = orthozer

    def nao(self) -> int:
        """Return the number of original atomic orbitals.

        Returns
        -------
        int
            Number of original atomic orbitals.
        """
        return self._orthozer.shape[-1]

    def convert2(self, mat: torch.Tensor) -> torch.Tensor:
        """
        Convert the last 2 dimensions of the matrix with shape (..., nao, nao)
        into the new orbital basis sets with shape (..., nao2, nao2)

        Parameters
        ----------
        mat: torch.Tensor
            Matrix to be converted.

        Returns
        -------
        torch.Tensor
            Converted matrix.
        """
        res = self._orthozer.transpose(-2, -1).conj() @ mat @ self._orthozer
        return res

    def convert4(self, mat: torch.Tensor) -> torch.Tensor:
        """
        Convert the last 4 dimensions of the matrix with shape (..., nao, nao, nao, nao)
        into the new orbital basis sets with shape (..., nao2, nao2, nao2, nao2).

        Parameters
        ----------
        mat: torch.Tensor
            Matrix to be converted.

        Returns
        -------
        torch.Tensor
            Converted matrix.
        """
        orthozer = self._orthozer
        res = torch.einsum("...ijkl,...im,...jn,...kp,...lq->...mnpq", mat,
                           orthozer, orthozer, orthozer, orthozer)
        return res

    def unconvert_dm(self, dm: torch.Tensor) -> torch.Tensor:
        """
        Convert back the density matrix from the density matrix in the new orbital
        basis with shape (..., nao2, nao2) to the original orbital basis
        (..., nao, nao)

        Parameters
        ----------
        dm: torch.Tensor
            Density matrix in the new orbital basis.

        Returns
        -------
        torch.Tensor
            Density matrix in the original orbital basis.
        """
        dm = torch.einsum("...kl,ik,jl->ij", dm, self._orthozer,
                          self._orthozer.conj())
        return dm

    def getparamnames(self, methodname: str, prefix: str = "") -> List[str]:
        """Get the parameter names of the orbital orthogonalizer.

        Parameters
        ----------
        methodname: str
            Method name.
        prefix: str
            Prefix of the parameter names.

        Returns
        -------
        List[str]
            List of parameter names.
        """
        return [prefix + "_orthozer"]



from typing import List, Callable, NamedTuple
import torch


# list of tableaus
class _Tableau(NamedTuple):
    """To specify a particular method, one needs to provide the integer s
    (the number of stages), and the coefficients a[i,j] (for 1 ≤ j < i ≤ s),
    b[i] (for i = 1, 2, ..., s) and c[i] (for i = 2, 3, ..., s). The matrix
    [aij] is called the Runge–Kutta matrix, while the b[i] and c[i] are known
    as the weights and the nodes. These data are usually arranged in a
    mnemonic device, known as a Butcher tableau (after John C. Butcher):

    Examples
    --------
    >>> euler = _Tableau(c=[0.0],
    ...                  b=[1.0],
    ...                  a=[[0.0]]
    ... )
    >>> euler.c
    [0.0]

    Attributes
    ----------
    c: List[float]
        The nodes
    b: List[float]
        The weights
    a: List[List[float]]
        The Runge-Kutta matrix

    """
    c: List[float]
    b: List[float]
    a: List[List[float]]


rk4_tableau = _Tableau(c=[0.0, 0.5, 0.5, 1.0],
                       b=[1 / 6., 1 / 3., 1 / 3., 1 / 6.],
                       a=[[0.0, 0.0, 0.0, 0.0], [0.5, 0.0, 0.0, 0.0],
                          [0.0, 0.5, 0.0, 0.0], [0.0, 0.0, 1.0, 0.0]])
rk38_tableau = _Tableau(c=[0.0, 1 / 3, 2 / 3, 1.0],
                        b=[1 / 8, 3 / 8, 3 / 8, 1 / 8],
                        a=[[0.0, 0.0, 0.0, 0.0], [1 / 3, 0.0, 0.0, 0.0],
                           [-1 / 3, 1.0, 0.0, 0.0], [1.0, -1.0, 1.0, 0.0]])
fwd_euler_tableau = _Tableau(c=[0.0], b=[1.0], a=[[0.0]])
midpoint_fableau = _Tableau(c=[0.0, 0.5],
                            b=[0.0, 1.0],
                            a=[[0.0, 0.0], [0.5, 0.0]])


def explicit_rk(tableau: _Tableau,
                fcn: Callable[..., torch.Tensor],
                y0: torch.Tensor,
                t: torch.Tensor,
                params: torch.Tensor,
                batch_size: int = 1,
                device="cpu"):
    """The family of explicit Runge–Kutta methods is a generalization
    of the RK4 method mentioned above.

    Examples
    --------
    >>> def lotka_volterra(t, y, params):
    ...     X, Y = y
    ...     alpha, beta, delta, gamma = params
    ...     dx_dt = alpha * X - beta * X * Y
    ...     dy_dt = delta * X * Y - gamma * Y
    ...     return torch.stack([dx_dt, dy_dt])
    >>> t = torch.linspace(0, 50, 100)
    >>> y_init = torch.rand(2, 1)
    >>> solver_param = [rk4_tableau,
    ...                 lotka_volterra,
    ...                 y_init,
    ...                 t,
    ...                 torch.tensor([1.1, 0.4, 0.1, 0.4])]
    >>> solution = explicit_rk(*solver_param)
    >>> print(solution.shape)
    torch.Size([100, 2, 1])

    For solving multiple ODEs, we can feed in multiple initial conditions and parameters
    >>> def lotka_volterra(t, y, params):
    ...     X, Y = y
    ...     alpha, beta, delta, gamma = params
    ...     dx_dt = alpha * X - beta * X * Y
    ...     dy_dt = delta * X * Y - gamma * Y
    ...     return torch.stack([dx_dt, dy_dt])
    >>> t = torch.linspace(0, 50, 100)
    >>> batch_size = 10
    >>> y_init = torch.rand(2, batch_size)
    >>> params = torch.rand(4, batch_size)
    >>> solver_param = [rk4_tableau,
    ...                 lotka_volterra,
    ...                 y_init,
    ...                 t,
    ...                 params]
    >>> solution = explicit_rk(*solver_param, batch_size=batch_size, device='cpu')
    >>> print(solution.shape)
    torch.Size([100, 2, 10])

    Parameters
    ----------
    fcn: callable dy/dt = fcn(t, y, *params)
        The function to be integrated. It should produce output of list of
        tensors following the shapes of tuple `y`. `t` should be a single element.
    t: torch.Tensor (nt,)
        The time values to integrate over. Can be generated by `linspace` (Refer example above).
    y0: list of torch.Tensor (*ny)
        The list of initial values.
    params: list
        List of input parameters for the function `fcn`.
    batch_size: int
        The batch size to compute the RK method. Default is 1.
    device: str
        The device to compute the RK method. Default is "cpu".

    Returns
    -------
    yt: list of torch.Tensor (nt, *ny, batch_size)
        The value of `y` at the given time `t` for each batch.

    Notes
    -----
    We require the function `fcn` to have t as the first argument, y as the second argument and
    the rest of the arguments as `params`. The function should return a tensor of the same shape as `y`.
    Even though t may not be used in the function, it is required to have it as the first argument. For
    multiple derivatives, pass the y as a list of tensors. If multiple parameters are required,
    they should be passed as a list. Refer example above.

    References
    ----------
    [1].. https://en.wikipedia.org/wiki/Runge%E2%80%93Kutta_methods#Explicit_Runge.E2.80.93Kutta_methods

    """

    assert y0.shape[
        -1] == batch_size, "Number of initial conditions should match the batch size"

    c, a, b = torch.tensor(tableau.c, device=device), torch.tensor(
        tableau.a, device=device), torch.tensor(tableau.b, device=device)
    t = torch.tensor(t).clone().detach().to(device)
    y0 = torch.tensor(y0).clone().detach().to(device)
    if params is not None:
        params = torch.tensor(params).to(device)
    if len(y0.shape) == 1:
        y0 = y0.unsqueeze(0)
    s = len(c)
    num_steps = len(t)
    yt_list = [y0]
    y = y0.clone()
    h = (t[-1] - t[0]) / (num_steps - 1)

    for i in range(1, num_steps):
        t0 = t[i]
        k = torch.zeros((s,) + y.shape, device=device)

        # Vectorize stage computation by stacking inputs
        for i in range(s):
            y_sum = y + h * torch.sum(
                a[i, :i].unsqueeze(-1).unsqueeze(-1) * k[:i], dim=0)
            if len(y0.shape) == 1:
                y_sum = y_sum.squeeze(0)
            k[i] = fcn(t=t0 + c[i] * h, y=y_sum.to(device), params=params)
        y = y + h * torch.sum(b.unsqueeze(-1).unsqueeze(-1) * k, dim=0)
        yt_list.append(y)

    yt = torch.stack(yt_list)
    return yt


# list of methods
def rk38_ivp(fcn: Callable[..., torch.Tensor],
             y0: torch.Tensor,
             t: torch.Tensor,
             params: torch.Tensor,
             batch_size: int = 1,
             device="cpu",
             **kwargs):
    """A slight variation of "the" Runge–Kutta method is also due to
    Kutta in 1901 and is called the 3/8-rule.[19] The primary advantage
    this method has is that almost all of the error coefficients are
    smaller than in the popular method, but it requires slightly more
    FLOPs (floating-point operations) per time step.

    Examples
    --------
    >>> def lotka_volterra(t, y, params):
    ...     X, Y = y
    ...     alpha, beta, delta, gamma = params
    ...     dx_dt = alpha * X - beta * X * Y
    ...     dy_dt = delta * X * Y - gamma * Y
    ...     return torch.stack([dx_dt, dy_dt])
    >>> t = torch.linspace(0, 50, 100)
    >>> y_init = torch.rand(2, 1)
    >>> solver_param = [lotka_volterra,
    ...                 y_init,
    ...                 t,
    ...                 torch.tensor([1.1, 0.4, 0.1, 0.4])]
    >>> solution = rk38_ivp(*solver_param)
    >>> print(solution.shape)
    torch.Size([100, 2, 1])

    Parameters
    ----------
    fcn: callable dy/dt = fcn(t, y, *params)
        The function to be integrated. It should produce output of list of
        tensors following the shapes of tuple `y`. `t` should be a single element.
    t: torch.Tensor (nt,)
        The integrated times
    y0: list of torch.Tensor (*ny)
        The list of initial values
    params: list
        List of any other parameters
    batch_size: int
        The batch size to compute the RK method. Default is 1.
    device: str
        The device to compute the RK method. Default is "cpu".
    **kwargs: dict
        Any other keyword arguments

    Returns
    -------
    yt: list of torch.Tensor (nt,*ny)
        The value of `y` at the given time `t`

    """
    return explicit_rk(rk38_tableau, fcn, y0, t, params, batch_size, device)


def fwd_euler_ivp(fcn: Callable[..., torch.Tensor],
                  y0: torch.Tensor,
                  t: torch.Tensor,
                  params: torch.Tensor,
                  batch_size: int = 1,
                  device="cpu",
                  **kwargs):
    """However, the simplest Runge–Kutta method is the (forward) Euler method,
    given by the formula $y_{n+1} = y_{n} + hf(t_{n}, y_{n}). This is the only
    consistent explicit Runge–Kutta method with one stage.

    Examples
    --------
    >>> def lotka_volterra(t, y, params):
    ...     X, Y = y
    ...     alpha, beta, delta, gamma = params
    ...     dx_dt = alpha * X - beta * X * Y
    ...     dy_dt = delta * X * Y - gamma * Y
    ...     return torch.stack([dx_dt, dy_dt])
    >>> t = torch.linspace(0, 50, 1000)
    >>> y_init = torch.randn(2, 1)
    >>> solver_param = [lotka_volterra,
    ...                 y_init,
    ...                 t,
    ...                 torch.tensor([1.1, 0.4, 0.1, 0.4])]
    >>> solution = fwd_euler_ivp(*solver_param)
    >>> print(solution.shape)
    torch.Size([1000, 2, 1])

    Parameters
    ----------
    fcn: callable dy/dt = fcn(t, y, *params)
        The function to be integrated. It should produce output of list of
        tensors following the shapes of tuple `y`. `t` should be a single element.
    t: torch.Tensor (nt,)
        The integrated times
    y0: list of torch.Tensor (*ny)
        The list of initial values
    params: list
        List of any other parameters
    batch_size: int
        The batch size to compute the RK method. Default is 1.
    device: str
        The device to compute the RK method. Default is "cpu".
    **kwargs: dict
        Any other keyword arguments

    Returns
    -------
    yt: list of torch.Tensor (nt,*ny)
        The value of `y` at the given time `t`

    """
    return explicit_rk(fwd_euler_tableau, fcn, y0, t, params, batch_size,
                       device)


def rk4_ivp(fcn: Callable[..., torch.Tensor],
            y0: torch.Tensor,
            t: torch.Tensor,
            params: torch.Tensor,
            batch_size: int = 1,
            device="cpu",
            **kwargs):
    """The most commonly used Runge Kutta method to find the solution
    of a differential equation is the RK4 method, i.e., the fourth-order
    Runge-Kutta method. The Runge-Kutta method provides the approximate
    value of y for a given point x. Only the first order ODEs can be
    solved using the Runge Kutta RK4 method.

    Examples
    --------
    >>> def lotka_volterra(t, y, params):
    ...     X, Y = y
    ...     alpha, beta, delta, gamma = params
    ...     dx_dt = alpha * X - beta * X * Y
    ...     dy_dt = delta * X * Y - gamma * Y
    ...     return torch.stack([dx_dt, dy_dt])
    >>> t = torch.linspace(0, 50, 100)
    >>> y_init = torch.rand(2, 1)
    >>> solver_param = [lotka_volterra,
    ...                 y_init,
    ...                 t,
    ...                 torch.tensor([1.1, 0.4, 0.1, 0.4])]
    >>> solution = rk4_ivp(*solver_param)
    >>> print(solution.shape)
    torch.Size([100, 2, 1])

    Parameters
    ----------
    fcn: callable dy/dt = fcn(t, y, *params)
        The function to be integrated. It should produce output of list of
        tensors following the shapes of tuple `y`. `t` should be a single element.
    t: torch.Tensor (nt,)
        The integrated times
    y0: list of torch.Tensor (*ny)
        The list of initial values
    params: list
        List of any other parameters
    batch_size: int
        The batch size to compute the RK method. Default is 1.
    device: str
        The device to compute the RK method. Default is "cpu".
    **kwargs: dict
        Any other keyword arguments

    Returns
    -------
    yt: list of torch.Tensor (nt,*ny)
        The value of `y` at the given time `t`

    """
    return explicit_rk(rk4_tableau, fcn, y0, t, params, batch_size, device)


def mid_point_ivp(fcn: Callable[..., torch.Tensor],
                  y0: torch.Tensor,
                  t: torch.Tensor,
                  params: torch.Tensor,
                  batch_size: int = 1,
                  device="cpu",
                  **kwargs):
    """The explicit midpoint method is sometimes also known as the
    modified Euler method, the implicit method is the most simple
    collocation method, and, applied to Hamiltonian dynamics, a
    symplectic integrator.

    Examples
    --------
    >>> def lotka_volterra(t, y, params):
    ...     X, Y = y
    ...     alpha, beta, delta, gamma = params
    ...     dx_dt = alpha * X - beta * X * Y
    ...     dy_dt = delta * X * Y - gamma * Y
    ...     return torch.stack([dx_dt, dy_dt])
    >>> t = torch.linspace(0, 50, 100)
    >>> y_init = torch.rand(2, 1)
    >>> solver_param = [lotka_volterra,
    ...                 y_init,
    ...                 t,
    ...                 torch.tensor([1.1, 0.4, 0.1, 0.4])]
    >>> solution = rk4_ivp(*solver_param)
    >>> print(solution.shape)
    torch.Size([100, 2, 1])

    Parameters
    ----------
    fcn: callable dy/dt = fcn(t, y, *params)
        The function to be integrated. It should produce output of list of
        tensors following the shapes of tuple `y`. `t` should be a single element.
    t: torch.Tensor (nt,)
        The integrated times
    y0: list of torch.Tensor (*ny)
        The list of initial values
    params: list
        List of any other parameters
    batch_size: int
        The batch size to compute the RK method. Default is 1.
    device: str
        The device to compute the RK method. Default is "cpu".
    **kwargs: dict
        Any other keyword arguments

    Returns
    -------
    yt: list of torch.Tensor (nt,*ny)
        The value of `y` at the given time `t`

    """
    return explicit_rk(midpoint_fableau, fcn, y0, t, params, batch_size, device)



"""
Featurizer for SE(3)-equivariant Graph Neural Networks.
"""

import numpy as np
from deepchem.feat.graph_data import GraphData
from deepchem.feat import MolecularFeaturizer
from rdkit import Chem
from rdkit.Chem import AllChem
import logging
from typing import List, Tuple, Optional, Any
from deepchem.utils.typing import RDKitMol, ArrayLike

logger = logging.getLogger(__name__)


class EquivariantGraphFeaturizer(MolecularFeaturizer):
    """Featurizer for Equivariant Graph Neural Networks.

    This featurizer constructs graph representations of molecular structures,
    capturing atomic features, pairwise distances, and spatial positions. These
    features are tailored for use in Equivariant models with QM9 dataset
    as described in [1]_.

    Features include:
    - **Node features**: Atomic one-hot encodings and additional descriptors.
    - **Edge features**: Vector displacements between atom pairs.
    - **Edge weights**: Discretized pairwise distances in one-hot encoding.
    - **Atomic coordinates**: 3D positions of atoms.

    Examples
    --------
    >>> from rdkit import Chem
    >>> import deepchem as dc
    >>> mol = Chem.MolFromSmiles('CCO')
    >>> featurizer = dc.feat.EquivariantGraphFeaturizer(fully_connected=True, embeded=True)
    >>> features = featurizer.featurize([mol])
    >>> type(features[0])
    <class 'deepchem.feat.graph_data.GraphData'>
    >>> features[0].node_features.shape  # (N, F)
    (3, 6)

    Notes
    -----
    This class requires RDKit to be installed.

    References
    ----------
    .. [1] Fuchs, F. B., et al. "SE(3)-Transformers: 3D Roto-Translation Equivariant
           Attention Networks." arXiv preprint arXiv:2006.10503 (2020).
    """

    def __init__(self,
                 fully_connected: bool = False,
                 weight_bins: Optional[List[Any]] = None,
                 embeded: bool = False):
        """
        Parameters
        ----------
        fully_connected : bool, optional (default False)
            If True, generates fully connected graphs with distance-based edges.
        weight_bins : list of float, optional (default [1.0, 2.0, 3.0, 4.0])
            Bin boundaries for discretizing edge weights.
        embeded : bool, optional (default False)
            Whether to embed 3D coordinates using RDKit's ETKDG method.
        """
        self.fully_connected = fully_connected
        self.embeded = embeded
        self.weight_bins = weight_bins if weight_bins is not None else [
            1.0, 2.0, 3.0, 4.0
        ]

    def _featurize(self, datapoint: RDKitMol, **kwargs) -> GraphData:
        """Featurizes molecules into GraphData objects.

        Parameters
        ----------
        datapoint : rdkit.Chem.rdchem.Mol
            RDKit Mol object.

        Returns
        -------
        np.ndarray of GraphData
            List of GraphData objects for the molecule.
        """

        if self.embeded:
            datapoint = Chem.AddHs(datapoint)
            AllChem.EmbedMolecule(datapoint, AllChem.ETKDG())
            datapoint = Chem.RemoveHs(datapoint)

        node_features, positions = self._get_node_features(datapoint)
        positions = np.array(positions, dtype=np.float32)

        if self.fully_connected:
            src, dst, edge_features, edge_weights = self._get_fully_connected_edges(
                positions)
        else:
            src, dst, edge_features, edge_weights = self._get_bonded_edges(
                datapoint, positions)

        edge_weights = self._discretize_and_one_hot(edge_weights)
        node_features = np.array(node_features, dtype=np.float32)
        edge_features = np.array(edge_features, dtype=np.float32)

        return GraphData(node_features=node_features,
                         edge_index=np.array([src, dst]),
                         edge_features=edge_features,
                         edge_weights=edge_weights,
                         positions=positions)

    def _get_node_features(self, mol: RDKitMol) -> Tuple[ArrayLike, ArrayLike]:
        """Generates node features and positions.

        Parameters
        ----------
        mol : rdkit.Chem.rdchem.Mol
            RDKit Mol object.

        Returns
        -------
        tuple
            A tuple of node features (np.ndarray) and positions (np.ndarray).
        """
        atom_features, positions = [], []

        for atom in mol.GetAtoms():
            atomic_number = atom.GetAtomicNum()
            one_hot = self._one_hot(atomic_number,
                                    [1, 6, 7, 8, 9])  # H, C, N, O, F
            additional_features = [atomic_number]
            atom_features.append(one_hot + additional_features)
            pos = mol.GetConformer().GetAtomPosition(atom.GetIdx())
            positions.append([pos.x, pos.y, pos.z])

        return atom_features, positions

    def _get_bonded_edges(
        self, mol: RDKitMol, positions: np.ndarray
    ) -> Tuple[ArrayLike, ArrayLike, ArrayLike, ArrayLike]:
        """Generates edges based on bonds.

        Parameters
        ----------
        mol : rdkit.Chem.rdchem.Mol
            RDKit Mol object.
        positions : np.ndarray
            Atomic positions.

        Returns
        -------
        tuple
            A tuple of source indices, destination indices, edge features, and edge weights.
        """
        src, dst, edge_features, edge_weights = [], [], [], []

        for bond in mol.GetBonds():
            i, j = bond.GetBeginAtomIdx(), bond.GetEndAtomIdx()
            src += [i, j]
            dst += [j, i]

            dist = positions[j] - positions[i]
            edge_features.append(dist)
            edge_features.append(-dist)
            edge_weights.append([np.linalg.norm(dist)])
            edge_weights.append([np.linalg.norm(dist)])

        return np.array(src), np.array(dst), np.array(edge_features), np.array(
            edge_weights)

    def _get_fully_connected_edges(
        self, positions: np.ndarray
    ) -> Tuple[ArrayLike, ArrayLike, ArrayLike, ArrayLike]:
        """Generates fully connected edges.

        Parameters
        ----------
        positions : np.ndarray
            Atomic positions.

        Returns
        -------
        tuple
            A tuple of source indices, destination indices, edge features, and edge weights.
        """
        num_atoms = len(positions)
        src, dst, edge_features, edge_weights = [], [], [], []

        for i in range(num_atoms):
            for j in range(num_atoms):
                if i != j:
                    src.append(i)
                    dst.append(j)
                    dist = positions[j] - positions[i]
                    edge_features.append(dist)
                    edge_weights.append([np.linalg.norm(dist)])

        return np.array(src), np.array(dst), np.array(edge_features), np.array(
            edge_weights)

    def _discretize_and_one_hot(self, edge_weights: ArrayLike) -> np.ndarray:
        """Discretizes edge weights into bins and converts to one-hot encoding.

        Parameters
        ----------
        edge_weights : np.ndarray
            Continuous edge weights to be discretized.

        Returns
        -------
        np.ndarray
            One-hot encoded edge weights.
        """
        edge_weights = np.array(edge_weights).flatten()
        digitized = np.digitize(edge_weights, self.weight_bins, right=False)
        num_bins = len(self.weight_bins) + 1
        one_hot_weights = np.zeros((len(digitized), num_bins))
        one_hot_weights[np.arange(len(digitized)), digitized] = 1
        return one_hot_weights

    def _one_hot(self, value: int, allowable_set: List) -> List[int]:
        """Generates a one-hot encoding for a value.

        Parameters
        ----------
        value : int
            Value to encode.
        allowable_set : List
            List of allowable values.

        Returns
        -------
        np.ndarray
            One-hot encoded vector.
        """
        return [1 if value == x else 0 for x in allowable_set]



"""
Tox21 dataset loader.
"""
import os
import deepchem as dc
from deepchem.molnet.load_function.molnet_loader import TransformerGenerator, _MolnetLoader
from deepchem.data import Dataset
from typing import List, Optional, Tuple, Union

TOX21_URL = "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/tox21.csv.gz"
TOX21_TASKS = [
    'NR-AR', 'NR-AR-LBD', 'NR-AhR', 'NR-Aromatase', 'NR-ER', 'NR-ER-LBD',
    'NR-PPAR-gamma', 'SR-ARE', 'SR-ATAD5', 'SR-HSE', 'SR-MMP', 'SR-p53'
]


class _Tox21Loader(_MolnetLoader):

    def create_dataset(self) -> Dataset:
        dataset_file = os.path.join(self.data_dir, "tox21.csv.gz")
        if not os.path.exists(dataset_file):
            dc.utils.data_utils.download_url(url=TOX21_URL,
                                             dest_dir=self.data_dir)
        loader = dc.data.CSVLoader(tasks=self.tasks,
                                   feature_field="smiles",
                                   featurizer=self.featurizer)
        return loader.create_dataset(dataset_file, shard_size=8192)


def load_tox21(
    featurizer: Union[dc.feat.Featurizer, str] = 'ECFP',
    splitter: Union[dc.splits.Splitter, str, None] = 'scaffold',
    transformers: List[Union[TransformerGenerator, str]] = ['balancing'],
    reload: bool = True,
    data_dir: Optional[str] = None,
    save_dir: Optional[str] = None,
    tasks: List[str] = TOX21_TASKS,
    **kwargs
) -> Tuple[List[str], Tuple[Dataset, ...], List[dc.trans.Transformer]]:
    """Load Tox21 dataset

    The "Toxicology in the 21st Century" (Tox21) initiative created a public
    database measuring toxicity of compounds, which has been used in the 2014
    Tox21 Data Challenge. This dataset contains qualitative toxicity measurements
    for 8k compounds on 12 different targets, including nuclear receptors and
    stress response pathways.

    Random splitting is recommended for this dataset.

    The raw data csv file contains columns below:

    - "smiles" - SMILES representation of the molecular structure
    - "NR-XXX" - Nuclear receptor signaling bioassays results
    - "SR-XXX" - Stress response bioassays results

    please refer to https://tripod.nih.gov/tox21/challenge/data.jsp for details.

    Parameters
    ----------
    featurizer: Featurizer or str
        the featurizer to use for processing the data.  Alternatively you can pass
        one of the names from dc.molnet.featurizers as a shortcut.
    splitter: Splitter or str
        the splitter to use for splitting the data into training, validation, and
        test sets.  Alternatively you can pass one of the names from
        dc.molnet.splitters as a shortcut.  If this is None, all the data
        will be included in a single dataset.
    transformers: list of TransformerGenerators or strings
        the Transformers to apply to the data.  Each one is specified by a
        TransformerGenerator or, as a shortcut, one of the names from
        dc.molnet.transformers.
    reload: bool
        if True, the first call for a particular featurizer and splitter will cache
        the datasets to disk, and subsequent calls will reload the cached datasets.
    data_dir: str
        a directory to save the raw data in
    save_dir: str
        a directory to save the dataset in
    tasks: List[str], (optional)
        Specify the set of tasks to load. If no task is specified, then it loads
    the default set of tasks which are NR-AR, NR-AR-LBD, NR-AhR, NR-Aromatase, NR-ER,
    NR-ER-LBD, NR-PPAR-gamma, SR-ARE, SR-ATAD5, SR-HSE, SR-MMP, SR-p53.

    References
    ----------
    .. [1] Tox21 Challenge. https://tripod.nih.gov/tox21/challenge/
    """
    loader = _Tox21Loader(featurizer, splitter, transformers, tasks, data_dir,
                          save_dir, **kwargs)
    return loader.load_dataset('tox21', reload)
